/*
 * D-Pixed grid format (*.dgr) plug-in for GIMP
 *
 * Copyright (C) 1999, 2000  Masahiro SAKAI
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "dpixplugins-conf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimp/gimpintl.h>
#include "dpform.h"
#include "dpixplugins-intl.h"

/* Declare local functions.
 */
static void query(void);
static void run(gchar* name,
                gint nparams,
                GimpParam* param,
                gint* nreturn_vals, GimpParam** return_vals);
static gboolean dgr_export(gint32 imageID, const char* filename);
static gboolean dgr_import(gint32 imageID, const char* filename);
static gboolean filename_dialog(const char* title);

/***** Local vars *****/

GimpPlugInInfo PLUG_IN_INFO = {
    NULL,                       /* init_proc */
    NULL,                       /* quit_proc */
    query,                      /* query_proc */
    run                         /* run_proc */
};

static char* filename = NULL;

/***** Functions *****/

MAIN()

static void query(void)
{
    static GimpParamDef dgr_export_args [] =
    {
        {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
        {GIMP_PDB_IMAGE, "image", "Input image"},
        {GIMP_PDB_DRAWABLE, "drawable", "Input drawable (not used)"},
        {GIMP_PDB_STRING, "output_filename", "DGR filename."}
    };
    static gint dgr_export_nargs = sizeof (dgr_export_args) / sizeof (dgr_export_args[0]);

    static GimpParamDef dgr_import_args [] =
    {
        {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
        {GIMP_PDB_IMAGE, "image", "Input image"},
        {GIMP_PDB_DRAWABLE, "drawable", "Input drawable (not used)"},
        {GIMP_PDB_STRING, "output_filename", "DGR filename."}
    };
    static gint dgr_import_nargs = sizeof (dgr_export_args) / sizeof (dgr_export_args[0]);

/*    gimp_plugin_domain_register("gimp-dpix-plugins", LOCALEDIR); */

    gimp_install_procedure ("plug_in_dgr_export",
                            "Export guides as DGR(D-Pixed Grid) format.",
                            "",
                            "Masahiro SAKAI <sakai@tom.sfc.keio.ac.jp>",
                            "Masahiro SAKAI",
                            "2000",
                            N_("<Image>/Guide/Export to DGR file..."),
                            "INDEXED*,RGB*,GRAY*",
                            GIMP_PLUGIN,
                            dgr_export_nargs, 0,
                            dgr_export_args, NULL);

    gimp_install_procedure ("plug_in_dgr_import",
                            "Import guides from DGR(D-Pixed Grid) format.",
                            "",
                            "Masahiro SAKAI <sakai@tom.sfc.keio.ac.jp>",
                            "Masahiro SAKAI",
                            "2000",
                            N_("<Image>/Guide/Import from DGR file..."),
                            "INDEXED*,RGB*,GRAY*",
                            GIMP_PLUGIN,
                            dgr_import_nargs, 0,
                            dgr_import_args, NULL);
}


static
void run(gchar* name, gint nparams, GimpParam* param,
         gint* nreturn_vals, GimpParam** return_vals)
{
    static GimpParam values[1];
    GimpRunModeType run_mode;
    gint32 imageID;

    run_mode = (GimpRunModeType)param[0].data.d_int32;

    *nreturn_vals = 1;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = GIMP_PDB_SUCCESS;

    switch (run_mode) {
    case GIMP_RUN_INTERACTIVE:
    case GIMP_RUN_WITH_LAST_VALS:
        /*  Possibly retrieve data  */
        filename = (char*)g_malloc(gimp_get_data_size(name));
        gimp_get_data(name, filename);
        break;

    case GIMP_RUN_NONINTERACTIVE:
        /*  Make sure all the arguments are there!  */
        if (nparams != 4)
            values[0].data.d_status = GIMP_PDB_CALLING_ERROR;
        filename = g_strdup(param[3].data.d_string);
        break;

    default:
        break;
    }
        
    imageID = param[1].data.d_int32;

    if (!strcmp(name, "plug_in_dgr_export")) {
        if (run_mode == GIMP_RUN_INTERACTIVE) {
            INIT_I18N_UI();
            if (!filename_dialog(_("Export to DGR file..."))) {
                values[0].data.d_status = GIMP_PDB_CALLING_ERROR;
                return;
            }
        }
        if (!dgr_export(imageID, filename))
            values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
    } else if (!strcmp(name, "plug_in_dgr_import")) {
        if (run_mode == GIMP_RUN_INTERACTIVE) {
            INIT_I18N_UI();
            if (!filename_dialog(_("Import from DGR file..."))) {
                values[0].data.d_status = GIMP_PDB_CALLING_ERROR;
                return;
            }
        }
        if (!dgr_import(imageID, filename))
            values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
    }
    if (run_mode == GIMP_RUN_INTERACTIVE
        && values[0].data.d_status == GIMP_PDB_SUCCESS)
    {
        gimp_set_data(name, filename, strlen(filename) + 1);
    }

    return;
}


static
gint guide_entry_compare(gconstpointer a, gconstpointer b)
{
    return GPOINTER_TO_UINT(a) - GPOINTER_TO_UINT(b);
}

static
gboolean dgr_export(gint32 imageID, const char* filename)
{
    FILE* stream;
    GList *h_guides, *v_guides;
    gint32 guideID;
    DPGRID_HEADER header;
    guint x_grid_count, y_grid_count;
    guint i;

    stream = fopen(filename, "wb");
    if (!stream)
        return FALSE;

    h_guides = NULL;
    v_guides = NULL;
    
    for (guideID = gimp_image_find_next_guide(imageID, 0);
         guideID != -1;
         guideID = gimp_image_find_next_guide(imageID, guideID))
    {
        guint data;
        data = gimp_image_get_guide_position(imageID, guideID);
        if (data == -1)
            break;
        switch (gimp_image_get_guide_orientation(imageID, guideID)) {
        case GIMP_HORIZONTAL:
            h_guides = g_list_append(h_guides, GUINT_TO_POINTER(data));
            break;
        case GIMP_VERTICAL:
            v_guides = g_list_append(v_guides, GUINT_TO_POINTER(data));
            break;
        default:
            break;
        }
    }

    h_guides = g_list_sort(g_list_first(h_guides), guide_entry_compare);
    v_guides = g_list_sort(g_list_first(v_guides), guide_entry_compare);

    x_grid_count = g_list_length(v_guides);
    y_grid_count = g_list_length(h_guides);

    memset(&header, 0, sizeof(header));
    sprintf(header.m_Comment,
            "DGR - D-Pixed grid file.\n"
            "Written by GIMP/%d.%d.%d with DGR-plugin/%s\n\n",
            gimp_major_version, gimp_minor_version, gimp_micro_version,
            VERSION);
    header.m_Width  = GUINT32_TO_LE(gimp_image_width(imageID));
    header.m_Height = GUINT32_TO_LE(gimp_image_height(imageID));
    header.m_XGrid  = GUINT32_TO_LE(x_grid_count);
    header.m_YGrid  = GUINT32_TO_LE(y_grid_count);
    header.m_Flags  = GUINT32_TO_LE(0);
    fwrite(&header, sizeof(header), 1, stream);

    for (i = 0; i < x_grid_count; i++) {
        guint16 tmp;
        tmp = GPOINTER_TO_UINT(g_list_nth_data(v_guides, i));
        tmp = GUINT16_TO_LE(tmp);
        fwrite(&tmp, sizeof(guint16), 1, stream);
    }
    for (i = 0; i < y_grid_count; i++) {
        guint16 tmp;
        tmp = GPOINTER_TO_UINT(g_list_nth_data(h_guides, i));
        tmp = GUINT16_TO_LE(tmp);
        fwrite(&tmp, sizeof(guint16), 1, stream);
    }

    g_list_free(h_guides);
    g_list_free(v_guides);

    fclose(stream);

    return TRUE;
}


static
gboolean dgr_import(gint32 imageID, const char* filename)
{
    FILE* stream;
    DPGRID_HEADER header;
    int i;
    guint16 guide_pos;

    stream = fopen(filename, "rb");
    if (!stream)
        return FALSE;

    fread(&header, sizeof(header), 1, stream);
    if (strncmp(header.m_Comment, "DGR", 3)) {
        fclose(stream);
        g_message(_("%s: %s isn't DGR file.\n"), "dgr", filename);
        return FALSE;
    }
        
    header.m_Width    = GUINT32_FROM_LE(header.m_Width);
    header.m_Height   = GUINT32_FROM_LE(header.m_Height);
    header.m_XGrid    = GUINT32_FROM_LE(header.m_XGrid);
    header.m_YGrid    = GUINT32_FROM_LE(header.m_YGrid);
    header.m_Flags    = GUINT32_FROM_LE(header.m_Flags);
    header.m_Reserved = GUINT32_FROM_LE(header.m_Reserved);

    for (i = 0; i < header.m_XGrid; i++) {
        fread(&guide_pos, sizeof(guide_pos), 1, stream);
        gimp_image_add_vguide(imageID, GUINT16_FROM_LE(guide_pos));
    }
    for (i = header.m_XGrid; i < header.m_XGrid + header.m_YGrid; i++) {
        fread(&guide_pos, sizeof(guide_pos), 1, stream);
        gimp_image_add_hguide(imageID, GUINT16_FROM_LE(guide_pos));
    }

    fclose(stream);

    return TRUE;
}


static gboolean exec_flag;

static
void filename_dialog_ok_callback(GtkWidget* widget, GtkFileSelection* fs)
{
    g_free(filename);
    filename = g_strdup(gtk_file_selection_get_filename(fs));
    exec_flag = TRUE;
    gtk_widget_destroy(GTK_WIDGET(fs));
}

static
void filename_dialog_cancel_callback(GtkWidget* widget, GtkFileSelection* fs)
{
    gtk_widget_destroy(GTK_WIDGET(fs));
}

static
gboolean filename_dialog(const char* title)
{
    GtkWidget* dialog;

    exec_flag = FALSE;

    gimp_ui_init("dgr", FALSE);

    dialog = gtk_file_selection_new ((char*)title);
    gtk_file_selection_set_filename (GTK_FILE_SELECTION(dialog),
                                     filename ? filename : "");
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
                       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

    gtk_signal_connect(
        GTK_OBJECT(GTK_FILE_SELECTION(dialog)->ok_button),
        "clicked",
        (GtkSignalFunc)filename_dialog_ok_callback,
        dialog);
    gtk_signal_connect(
        GTK_OBJECT(GTK_FILE_SELECTION(dialog)->cancel_button),
        "clicked",
        (GtkSignalFunc)filename_dialog_cancel_callback,
        dialog);

    gtk_widget_show(dialog);
    gtk_main();

    return exec_flag;
}
