/*
 * D-Pixed palette format (*.dpf) plug-in for GIMP
 *
 * Copyright (C) 1999, 2000  Masahiro SAKAI
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "dpixplugins-conf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimp/gimpintl.h>
#include "dpform.h"
#include "dpixplugins-intl.h"

/* Declare local functions.
 */
static void query(void);
static void run(gchar* name,
                gint nparams,
                GimpParam* param,
                gint* nreturn_vals, GimpParam** return_vals);
static gboolean palette_to_dpf(const char* input_filename,
                               const char* output_filename);
static gboolean dpf_to_palette(const char* input_filename,
                               const char* output_filename);
static gboolean filename_dialog(const char* title);

/***** Local vars *****/

GimpPlugInInfo PLUG_IN_INFO = {
    NULL,                       /* init_proc */
    NULL,                       /* quit_proc */
    query,                      /* query_proc */
    run                         /* run_proc */
};

static char* input_filename = NULL;
static char* output_filename = NULL;

/***** Functions *****/

MAIN()

static void query(void)
{
    static GimpParamDef dpf_export_args [] =
    {
        {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
        {GIMP_PDB_STRING, "input_filename",  "filename of gimp palette."},
        {GIMP_PDB_STRING, "output_filename", "filename of D-Pixed palette (*.dpf)"}
    };
    static gint dpf_export_nargs = sizeof (dpf_export_args) / sizeof (dpf_export_args[0]);

    static GimpParamDef dpf_import_args [] =
    {
        {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
        {GIMP_PDB_STRING, "input_filename", "filename of D-Pixed palette (*.dpf)"},
        {GIMP_PDB_STRING, "output_filename",  "filename of gimp palette."}
    };
    static gint dpf_import_nargs = sizeof (dpf_export_args) / sizeof (dpf_export_args[0]);

    gimp_plugin_domain_register("gimp-dpix-plugins", LOCALEDIR);

    gimp_install_procedure ("extension_dpf_export",
                            "Convert palette to DPF(D-Pixed Palatte) format.",
                            "",
                            "Masahiro SAKAI <sakai@tom.sfc.keio.ac.jp>",
                            "Masahiro SAKAI",
                            "2000",
                            N_("<Toolbox>/Xtns/Data Converter/Palette => D-Pixed palette (*.dpf)"),
                            "",
                            GIMP_EXTENSION,
                            dpf_export_nargs, 0,
                            dpf_export_args, NULL);

    gimp_install_procedure ("extension_dpf_import",
                            "Convert D-Pixed Palatte file(*.dpf).",
                            "",
                            "Masahiro SAKAI <sakai@tom.sfc.keio.ac.jp>",
                            "Masahiro SAKAI",
                            "2000",
                            N_("<Toolbox>/Xtns/Data Converter/D-Pixed palette (*.dpf) => Palette"),
                            "",
                            GIMP_EXTENSION,
                            dpf_import_nargs, 0,
                            dpf_import_args, NULL);
}


static
void run(gchar* name, gint nparams, GimpParam* param,
         gint* nreturn_vals, GimpParam** return_vals)
{
    static GimpParam values[1];
    GimpRunModeType run_mode;
    gint32 imageID;

    run_mode = (GimpRunModeType)param[0].data.d_int32;

    *nreturn_vals = 1;
    *return_vals  = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = GIMP_PDB_SUCCESS;

    switch (run_mode) {
    case GIMP_RUN_INTERACTIVE:
    case GIMP_RUN_WITH_LAST_VALS:
        {
        /*  Possibly retrieve data  */
        char* tmp;
        gint size1, size2;
            
        if (gimp_get_data_size(name) < 2)
            break;
        
        tmp = (char*)g_malloc(gimp_get_data_size(name));
        gimp_get_data(name, tmp);
        
        size1 = strlen(tmp) + 1;
        size2 = strlen(tmp + size1) + 1;
        
        g_free(input_filename);
        g_free(output_filename);
        input_filename  = (char*)g_malloc(size1);
        output_filename = (char*)g_malloc(size2);
        memcpy(input_filename, tmp, size1);
        memcpy(output_filename, tmp + size1, size2);
        break;
        }
    case GIMP_RUN_NONINTERACTIVE:
        /*  Make sure all the arguments are there!  */
        if (nparams != 3)
            values[0].data.d_status = GIMP_PDB_CALLING_ERROR;
        g_free(input_filename);
        g_free(output_filename);
        input_filename  = g_strdup(param[1].data.d_string);
        output_filename = g_strdup(param[2].data.d_string);
        break;

    default:
        break;
    }

    imageID = param[1].data.d_int32;

    if (!strcmp(name, "extension_dpf_export")) {
        if (run_mode == GIMP_RUN_INTERACTIVE) {
            if (!filename_dialog(_("Palette => D-Pixed Palette (*.dpf)"))) {
                values[0].data.d_status = GIMP_PDB_CALLING_ERROR;
                return;
            }
        }
        if (!palette_to_dpf(input_filename, output_filename))
            values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
    } else if (!strcmp(name, "extension_dpf_import")) {
        if (run_mode == GIMP_RUN_INTERACTIVE) {
            if (!filename_dialog(_("D-Pixed Palette (*.dpf) => Palette"))) {
                values[0].data.d_status = GIMP_PDB_CALLING_ERROR;
                return;
            }
        }
        if (!dpf_to_palette(input_filename, output_filename))
            values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
    } else {
        values[0].data.d_status = GIMP_PDB_CALLING_ERROR;
        return;
    }

    if (run_mode == GIMP_RUN_INTERACTIVE
        && values[0].data.d_status == GIMP_PDB_SUCCESS)
    {
        gint size1, size2;
        char* tmp;
        size1 = strlen(input_filename) + 1;
        size2 = strlen(output_filename) + 1;

        tmp = (char*)g_malloc(size1 + size2);
        memcpy(tmp, input_filename, size1);
        memcpy(tmp + size1, output_filename, size2);
        gimp_set_data(name, tmp, size1 + size2);
        g_free(tmp);
    }

    return;
}


static
gboolean palette_to_dpf_impl(const char* filename, FILE* input, FILE* output)
{
    gchar  str[512];
    gchar* tok;
    gint   r, g, b;
    gint   linenum;
    GList* palette;
    
    r = g = b = 0;

    linenum = 0;
    palette = NULL;

    if (!fgets (str, 512, input)) {
        g_message (_("Loading palette %s:\nRead error"), filename);
        return FALSE;
    }
    if (strcmp(str, "GIMP Palette\n") && strcmp(str, "GIMP Palette\r\n")) {
        g_message (_("Loading palette %s:\n"
                         "Corrupt palette: missing magic header"), filename);
        return FALSE;
    }

    linenum++;

    while (!feof (input)) {
        if (!fgets (str, 512, input)) {
            if (feof (input))
                break;
            g_message (_("Loading palette %s (line %d):\nRead error"),
                       filename, linenum);
            return FALSE;
	}
        
        linenum++;

        if (str[0] != '#') {
            tok = strtok (str, " \t");
            if (tok)
                r = atoi (tok);
            else
                /* maybe we should just abort? */
                g_message (_("Loading palette %s (line %d):\n"
                             "Missing RED component"), filename, linenum);

            tok = strtok (NULL, " \t");
            if (tok)
                g = atoi (tok);
            else
                g_message (_("Loading palette %s (line %d):\n"
                             "Missing GREEN component"), filename, linenum);

            tok = strtok (NULL, " \t");
            if (tok)
                b = atoi (tok);
            else
                g_message (_("Loading palette %s (line %d):\n"
                             "Missing BLUE component"), filename, linenum);

            /* optional name */
            tok = strtok (NULL, "\n");

            if (r < 0 || r > 255 ||
                g < 0 || g > 255 ||
                b < 0 || b > 255)
                g_message (_("Loading palette %s (line %d):\n"
                             "RGB value out of range"), filename, linenum);

            {
                DP_RGBQuad* rgb = g_new(DP_RGBQuad, 1);
                rgb->rgbBlue     = b;
                rgb->rgbGreen    = g;
                rgb->rgbRed      = r;
                rgb->rgbReserved = 0;
                palette = g_list_append(palette, rgb);
            }
	}
    }

    {
        gint palette_count = g_list_length(palette);
        DPIXED_FILEPALETTEHEADER header;
        int i;

        memset(&header, 0, sizeof(header));
        sprintf(header.fileDescription,
                "PF\n"
                "DPF - D-Pixed palette file.\n"
                "Written by GIMP/%d.%d.%d with DPF-plugin/%s\n\n",
                gimp_major_version, gimp_minor_version, gimp_micro_version,
                VERSION);
        header.version  = GUINT32_TO_LE(0x00000100);
        header.size     = GUINT32_TO_LE(palette_count);
        fwrite(&header, sizeof(header), 1, output);

        for (i = 0; i < palette_count; i++)
            fwrite(g_list_nth_data(palette, i), sizeof(DP_RGBQuad), 1, output);
    }

    g_list_free(palette);

    return TRUE;
}


static
gboolean dpf_to_palette_impl(const char* filename, FILE* input, FILE* output)
{
    DPIXED_FILEPALETTEHEADER header;
    int i;

    fread(&header, sizeof(header), 1, input);
    if (strncmp(header.fileDescription, "PF", 2)) {
        g_message("%s is not D-Pixed palette file.", filename);
        return FALSE;
    }
    header.version  = GUINT32_FROM_LE(header.version);
    header.dwFlag   = GUINT32_FROM_LE(header.dwFlag);
    header.size     = GUINT32_FROM_LE(header.size);

    fprintf(output, "GIMP Palette\n");
    for (i = 0; i < header.size; i++) {
        DP_RGBQuad rgb;
        fread(&rgb, sizeof(rgb), 1, input);
        fprintf(output, "%3d %3d %3d\n",
                rgb.rgbRed, rgb.rgbGreen, rgb.rgbBlue);
    }

    return TRUE;
}


static
gboolean palette_to_dpf(const char* input_filename,
                        const char* output_filename)
{
    FILE* input  = NULL;
    FILE* output = NULL;
    gboolean result = FALSE;

    input  = fopen(input_filename, "rt");
    if (!input)
        goto finally;
    output = fopen(output_filename, "wb");
    if (!output)
        goto finally;

    result = palette_to_dpf_impl(input_filename, input, output);

finally:
    fclose(output);    
    fclose(input);
    return result;
}


static
gboolean dpf_to_palette(const char* input_filename,
                        const char* output_filename)
{
    FILE* input  = NULL;
    FILE* output = NULL;
    gboolean result = FALSE;

    input  = fopen(input_filename, "rb");
    if (!input)
        goto finally;
    output = fopen(output_filename, "wt");
    if (!output)
        goto finally;

    result = dpf_to_palette_impl(input_filename, input, output);

finally:
    fclose(output);    
    fclose(input);
    return result;
}


struct {
    GimpFileSelection* input_fileselection;
    GimpFileSelection* output_fileselection;
    gboolean exec_flag;
} interface;

static
void dialog_ok_callback(GtkWidget *widget, gpointer data)
{
    g_free(input_filename);
    g_free(output_filename);
    input_filename  = gimp_file_selection_get_filename(interface.input_fileselection);
    output_filename = gimp_file_selection_get_filename(interface.output_fileselection);
    interface.exec_flag = TRUE;
    gtk_widget_destroy(GTK_WIDGET(data));
}

static
gboolean filename_dialog(const char* title)
{
    GtkWidget* dlg;
    GtkWidget* frame;
    GtkWidget* table;
    GtkWidget* label;
    GtkWidget* file_selection;

    interface.exec_flag = FALSE;

    INIT_I18N_UI();
    gimp_ui_init("dgr", FALSE);

    dlg = gimp_dialog_new(title, "dpf",
                          gimp_standard_help_func, "",
                          GTK_WIN_POS_MOUSE,
                          FALSE, TRUE, FALSE,
                          _("OK"), dialog_ok_callback,
                          NULL, NULL, NULL, TRUE, FALSE,
                          _("Cancel"), gtk_widget_destroy,
                          NULL, 1, NULL, FALSE, TRUE, NULL);
    gtk_signal_connect(GTK_OBJECT(dlg), "destroy",
                       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

    /*  parameter settings  */
    frame = gtk_frame_new(_("Parameter Settings"));
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
    gtk_container_border_width(GTK_CONTAINER(frame), 6);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), frame, TRUE, TRUE, 0);

    table = gtk_table_new(2, 2, FALSE);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_container_border_width(GTK_CONTAINER(table), 4);
    gtk_container_add(GTK_CONTAINER(frame), table);

    label = gtk_label_new(_("Input file"));
    gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 0, 1);
    gtk_widget_show(label);
    file_selection = gimp_file_selection_new(
        _("Input file"),
        input_filename ? input_filename : "",
        FALSE, TRUE);
    gtk_table_attach_defaults(GTK_TABLE(table), file_selection, 1, 2, 0, 1);
    gtk_widget_show(file_selection);
    interface.input_fileselection = GIMP_FILE_SELECTION(file_selection);

    label = gtk_label_new(_("Output file"));
    gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 1, 2);
    gtk_widget_show(label);
    file_selection = gimp_file_selection_new(
        _("Output file"),
        output_filename ? output_filename : "",
        FALSE, TRUE/*FALSE*/); /* FIXME */
    gtk_table_attach_defaults(GTK_TABLE(table), file_selection, 1, 2, 1, 2);
    gtk_widget_show(file_selection);
    interface.output_fileselection = GIMP_FILE_SELECTION(file_selection);

    gtk_widget_show(frame);
    gtk_widget_show(table);
    gtk_widget_show(dlg);

    gtk_main();

    return interface.exec_flag;
}
