/*
 * D-Pixed palette file converter for GIMP
 *
 * Copyright (C) 1999, 2000  Masahiro SAKAI
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "dpixplugins-conf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include "getopt.h"
#include "dpform.h"
#include "intl.h"


static
gboolean convert(FILE* input, FILE* output)
{
    DPIXED_FILEPALETTEHEADER header;
    int i;

    fread(&header, sizeof(header), 1, input);
    if (strncmp(header.fileDescription, "PF", 2)) {
        fputs(_("Input is not D-Pixed Palette Format."), stderr);
        return FALSE;
    }
    header.version  = GUINT32_FROM_LE(header.version);
    header.dwFlag   = GUINT32_FROM_LE(header.dwFlag);
    header.size     = GUINT32_FROM_LE(header.size);

    fprintf(output, "GIMP Palette\n");
    for (i = 0; i < header.size; i++) {
        DP_RGBQuad rgb;
        fread(&rgb, sizeof(rgb), 1, input);
        fprintf(output, "%3d %3d %3d\n",
                rgb.rgbRed, rgb.rgbGreen, rgb.rgbBlue);
    }

    return TRUE;
}


/*
 *  Help and version information
 */
void usage (const char* progname) {
    fprintf(stderr, _("Usage: %s -o output,dpf filename\n"), progname);
}

const char* const help_message[] = {
    N_(" -o, --output=FILE              specify output file name as FILE"),
    N_(" -h, --help                     display this help message."),
    N_(" -v, --version                  display the version."),
    NULL
};

void
help(const char* progname)
{
    const char* const* p = help_message;

    fprintf(stderr, _("help: %s\n\n"), progname);
    usage(progname);
    while (*p) fprintf(stderr, "%s\n", _(*p++));
}

void version(const char* progname)
{
    fprintf(stderr, "version of %s: %s\n", progname, VERSION);
}


int main(int argc, char* argv[])
{
    static struct option longopts[] = {
        /* These options set a flag. */
        {"help",    0, NULL, 'h'},
        {"version", 0, NULL, 'v'},
        {"output",  1, NULL, 'o'},
        {0, 0, 0, 0}
    };
    int optc;
    char* progname = "dpf2gimpoal";
    /* char *progname = argv[0]; */
    char* output_filename = NULL;
    char* input_filename = NULL;

    while ((optc = getopt_long(argc, argv, "hvo:", longopts, NULL)) != EOF) {
        switch (optc) {
        case 'h':
            help(progname);
            exit(EXIT_SUCCESS);
            break;
        case 'v':
            version(progname);
            exit(EXIT_SUCCESS);
            break;
        case 'o':
            output_filename = optarg;
            break;
        default:
            usage(progname);
            exit(EXIT_FAILURE);
        }
    }
    
    /* Test whether we have a .dpf file name as argument.  */
    if (optind >= argc) {
        fprintf(stderr, _("no input file given"));
        exit(EXIT_FAILURE);
    }
    input_filename = argv[optind];
    
    {
        FILE* input  = NULL;
        FILE* output = NULL;
        int result   = EXIT_FAILURE;

        if (strcmp(input_filename, "-") == 0)
            input = stdin;
        else
            input = fopen(input_filename, "rb");
        
        if (!input) {
            fprintf(stderr, _("failed in opening input file\n"));
            goto finally;
        }

        if (strcmp(output_filename, "-") == 0)
            output = stdout;
        else
            output = fopen(output_filename, "wt");

        if (!output) {
            fprintf(stderr, _("failed in opening output file\n"));
            goto finally;
        }

        if (!convert(input, output))
            goto finally;

        result = EXIT_SUCCESS;

    finally:
        if (output && output != stdout)
            fclose(output);
        if (input && input != stdin)
            fclose(input);

        return result;
    }
}
