/*
 * This GIMP plug-in is designed to emulate D-Pixed layering dither
 *
 * Copyright (C) 1999, 2000  Masahiro SAKAI
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "dpixplugins-conf.h"

#include <stdio.h>
#include <stdlib.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimp/gimpintl.h>
#include "dpixplugins-intl.h"

/* Some useful macros */

#define PLUG_IN_NAME "plug_in_dpix_dissolve"
#define TILE_CACHE_SIZE  16
#define SCALE_WIDTH     125
#define ENTRY_WIDTH      50

typedef struct {
    gint opacity;
} DissolveValues;

typedef struct {
    gint run;
} DissolveInterface;


/* Declare local functions.
 */
static void query(void);
static void run(gchar *name,
                gint nparams,
                GimpParam *param,
                gint *nreturn_vals, GimpParam **return_vals);

static gint dissolve_dialog(void);
static void dissolve_ok_callback(GtkWidget *widget, gpointer data);

static void dissolve(GimpDrawable *drawable);


/***** Local vars *****/

GimpPlugInInfo PLUG_IN_INFO = {
    NULL,                       /* init_proc */
    NULL,                       /* quit_proc */
    query,                      /* query_proc */
    run                         /* run_proc */
};

static DissolveValues dvals = {
    16
};

static DissolveInterface dint = {
    FALSE                       /* run */
};


/***** Functions *****/

MAIN()

static void query(void)
{
    static GimpParamDef args[] = {
        {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
        {GIMP_PDB_IMAGE, "image", "Input image (unused)"},
        {GIMP_PDB_DRAWABLE, "drawable", "Input drawable"},
        {GIMP_PDB_INT32, "opacity", "Opacity (0..16)"}
    };
    static GimpParamDef *return_vals = NULL;
    static gint nargs = sizeof(args) / sizeof(args[0]);
    static gint nreturn_vals = 0;

    gimp_plugin_domain_register("gimp-dpix-plugins", LOCALEDIR);

    gimp_install_procedure(PLUG_IN_NAME,
                           "Emulate D-Pixed's layering",
                           "Emulate D-Pixed's layering with speficied opacity.",
                           "Masahiro SAKAI <sakai@tom.sfc.keio.ac.jp>",
                           "Masahiro SAKAI",
                           "2000",
                           N_("<Image>/Filters/Misc/D-Pixed Dissolve..."),
                           "RGBA, GRAYA, INDEXEDA",
                           GIMP_PLUGIN,
                           nargs, nreturn_vals, args, return_vals);
}


static
void run(gchar *name, gint nparams, GimpParam *param,
         gint *nreturn_vals, GimpParam **return_vals)
{
    static GimpParam values[1];
    GimpDrawable *drawable;
    GimpRunModeType run_mode;
    GimpPDBStatusType status = GIMP_PDB_SUCCESS;

    run_mode = (GimpRunModeType)param[0].data.d_int32;

    *nreturn_vals = 1;
    *return_vals = values;

    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;

    /*  Get the specified drawable  */
    drawable = gimp_drawable_get(param[2].data.d_drawable);

    INIT_LOCALE("gimp-std-plugins");

    switch (run_mode) {
    case GIMP_RUN_INTERACTIVE:
        /*  Possibly retrieve data  */
        gimp_get_data(PLUG_IN_NAME, &dvals);

        /*  First acquire information with a dialog  */
        if (!dissolve_dialog()) {
            gimp_drawable_detach(drawable);
            return;
        }
        break;

    case GIMP_RUN_NONINTERACTIVE:
        /*  Make sure all the arguments are there!  */
        if (nparams != 4)
            status = GIMP_PDB_CALLING_ERROR;
        if (status == GIMP_PDB_SUCCESS)
            dvals.opacity = param[3].data.d_int32;
        if ((status == GIMP_PDB_SUCCESS) && (dvals.opacity < 0) && (dvals.opacity > 16))
            status = GIMP_PDB_CALLING_ERROR;
        break;

    case GIMP_RUN_WITH_LAST_VALS:
        /*  Possibly retrieve data  */
        gimp_get_data(PLUG_IN_NAME, &dvals);
        break;

    default:
        break;
    }

    if (status == GIMP_PDB_SUCCESS) {
        /*  Make sure that the drawable has alpha channel */
        if (!(gimp_drawable_has_alpha(drawable->id))) {
            status = GIMP_PDB_EXECUTION_ERROR;
        } else {
            gimp_progress_init(_("Dissolveing..."));

            /*  set the tile cache size  */
            gimp_tile_cache_ntiles(TILE_CACHE_SIZE);

            /*  run the pixelize effect  */
            dissolve(drawable);

            if (run_mode != GIMP_RUN_NONINTERACTIVE)
                gimp_displays_flush();

            /*  Store data  */
            if (run_mode == GIMP_RUN_INTERACTIVE)
                gimp_set_data(PLUG_IN_NAME, &dvals, sizeof(DissolveValues));
        }
    }

    values[0].data.d_status = status;

    gimp_drawable_detach(drawable);
}



static
gint dissolve_dialog(void)
{
    GtkWidget* dlg;
    GtkWidget* frame;
    GtkWidget* table;
    GtkObject* adjustment;

    gtk_set_locale();
    gimp_ui_init("dpx_dissolve", FALSE);

    dlg = gimp_dialog_new(_("D-Pixed Dissolve"), "dissolve",
                          gimp_standard_help_func, "",
                          GTK_WIN_POS_MOUSE,
                          FALSE, TRUE, FALSE,
                          _("OK"), dissolve_ok_callback,
                          NULL, NULL, NULL, TRUE, FALSE,
                          _("Cancel"), gtk_widget_destroy,
                          NULL, 1, NULL, FALSE, TRUE, NULL);

    gtk_signal_connect(GTK_OBJECT(dlg), "destroy",
                       GTK_SIGNAL_FUNC(gtk_main_quit), NULL);

    /*  parameter settings  */
    frame = gtk_frame_new(_("Parameter Settings"));
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
    gtk_container_border_width(GTK_CONTAINER(frame), 6);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), frame, TRUE, TRUE, 0);

    table = gtk_table_new(1, 3, FALSE);
    gtk_table_set_col_spacings(GTK_TABLE(table), 4);
    gtk_container_border_width(GTK_CONTAINER(table), 4);
    gtk_container_add(GTK_CONTAINER(frame), table);

    adjustment = gimp_scale_entry_new(GTK_TABLE(table), 0, 0,
                                      _("Opacity:"),
                                      SCALE_WIDTH, ENTRY_WIDTH,
                                      dvals.opacity, 0, 16, 1, 8, 0,
                                      FALSE, 0, 16,
                                      NULL, NULL);
    gtk_signal_connect(GTK_OBJECT(adjustment), "value_changed",
                       GTK_SIGNAL_FUNC(gimp_int_adjustment_update),
                       &dvals.opacity);

    gtk_widget_show(frame);
    gtk_widget_show(table);
    gtk_widget_show(dlg);

    gtk_main();
    gdk_flush();

    return dint.run;
}

static
void dissolve_ok_callback(GtkWidget *widget, gpointer data)
{
    dint.run = TRUE;
    gtk_widget_destroy(GTK_WIDGET(data));
}


const guchar dither_tile[17][4][4] = {
    { {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0} },
    { {0, 0, 0, 1}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0} },
    { {0, 0, 0, 1}, {0, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 0, 0} },
    { {0, 1, 0, 1}, {0, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 0, 0} },
    { {0, 1, 0, 1}, {0, 0, 0, 0}, {0, 1, 0, 1}, {0, 0, 0, 0} },
    { {0, 1, 0, 1}, {0, 0, 1, 0}, {0, 1, 0, 1}, {0, 0, 0, 0} },
    { {0, 1, 0, 1}, {0, 0, 1, 0}, {0, 1, 0, 1}, {1, 0, 0, 0} },
    { {0, 1, 0, 1}, {1, 0, 1, 0}, {0, 1, 0, 1}, {1, 0, 0, 0} },
    { {0, 1, 0, 1}, {1, 0, 1, 0}, {0, 1, 0, 1}, {1, 0, 1, 0} },
    { {0, 1, 1, 1}, {1, 0, 1, 0}, {0, 1, 0, 1}, {1, 0, 1, 0} },
    { {0, 1, 1, 1}, {1, 0, 1, 0}, {1, 1, 0, 1}, {1, 0, 1, 0} },
    { {1, 1, 1, 1}, {1, 0, 1, 0}, {1, 1, 0, 1}, {1, 0, 1, 0} },
    { {1, 1, 1, 1}, {1, 0, 1, 0}, {1, 1, 1, 1}, {1, 0, 1, 0} },
    { {1, 1, 1, 1}, {1, 0, 1, 1}, {1, 1, 1, 1}, {1, 0, 1, 0} },
    { {1, 1, 1, 1}, {1, 0, 1, 1}, {1, 1, 1, 1}, {1, 1, 1, 0} },
    { {1, 1, 1, 1}, {1, 1, 1, 1}, {1, 1, 1, 1}, {1, 1, 1, 0} },
    { {1, 1, 1, 1}, {1, 1, 1, 1}, {1, 1, 1, 1}, {1, 1, 1, 1} }
};

static
void dissolve(GimpDrawable *drawable)
{
    GimpPixelRgn src_rgn, dest_rgn;
    guchar *src_row, *dest_row;
    gint x1, y1, x2, y2;
    gpointer pr;
    gint progress, max_progress;

    /* initialize */
    gimp_drawable_mask_bounds (drawable->id, &x1, &y1, &x2, &y2);
    gimp_pixel_rgn_init (&src_rgn, drawable,
                         x1, y1, (x2 - x1), (y2 - y1), FALSE, FALSE);
    gimp_pixel_rgn_init (&dest_rgn, drawable,
                         x1, y1, (x2 - x1), (y2 - y1), TRUE, TRUE);

    /* Initialize progress */
    progress = 0;
    max_progress = (x2 - x1) * (y2 - y1);

    for (pr = gimp_pixel_rgns_register (2, &src_rgn, &dest_rgn);
         pr != NULL;
         pr = gimp_pixel_rgns_process (pr))
    {
        guint row, col;
        src_row = src_rgn.data;
        dest_row = dest_rgn.data;
        
        for (row = 0; row < src_rgn.h; row++)
	{
            guchar *src, *dest;
            src = src_row;
            dest = dest_row;
            
            for (col = 0; col < src_rgn.w; col++)
            {
                gint x = src_rgn.x + col;
                gint y = src_rgn.y + row;
                guint b;

                for (b = 0; b < src_rgn.bpp - 1; b++)
                    dest[b] = src[b];

                if (dither_tile[dvals.opacity][y % 4][x % 4])
                    dest[dest_rgn.bpp - 1] = src[src_rgn.bpp - 1];
                else
                    dest[dest_rgn.bpp - 1] = 0;

                src += src_rgn.bpp;
                dest += dest_rgn.bpp;
	    }
            
            src_row += src_rgn.rowstride;
            dest_row += dest_rgn.rowstride;
	}

        /* Update progress */
        progress += src_rgn.w * src_rgn.h;
        gimp_progress_update ((double) progress / (double) max_progress);
    }

    /*  update the region  */
    gimp_drawable_flush (drawable);
    gimp_drawable_merge_shadow (drawable->id, TRUE);
    gimp_drawable_update (drawable->id, x1, y1, (x2 - x1), (y2 - y1));
}
