/*
 * D-Pixed image format (*.dpx) plug-in for GIMP
 *
 * Copyright (C) 1999, 2000  Masahiro SAKAI
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "dpixplugins-conf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <locale.h>
#ifdef HAVE_MMAP
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#endif
#ifdef HAVE_ICONV
#include <errno.h>
#include <assert.h>
#include <iconv.h>
#endif
#ifdef G_OS_WIN32
#include <windows.h>
#undef RGB
#else
#include <sys/stat.h>
#endif
#include <gdk/gdk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimp/gimpintl.h>
#include "dpform.h"
#include "dpixplugins-intl.h"

#define BIT(x) (1 << (x))

static const char* const PROG_NAME = "dpx";

/* 
 * Declare some local functions.
 */

static void query(void);
static void run(char* name,
                int nparams,
                GimpParam* param, int* nreturn_vals, GimpParam** return_vals);

static gint32 load_image(const char* filename);

static guint32 read_layer(const guchar* data, const guchar* buffer_end,
                          guint32 imageID);

static gboolean decompress_image(const guchar* src_data, guint src_size,
                                 guchar* dest_data, guint dest_size);

static guint read_mask(const DPIXED_FILEIMAGEHEADER* header,
                       const guchar* src,  guchar* dst,
                       guint width, guint height);

static guint decompress_mask(const guchar* src, guchar* dst,
                             guint decompressed_size_4bit);

static guint read_additional_data_field(gint imageID,
                                        const guchar* begin,
                                        const guchar* end);

static void read_grid(gint imageID, const guchar* begin, const guchar* end);
static void read_frame(gint imageID, const guchar* begin, const guchar* end);
static void read_shape(gint imageID, const guchar* begin, const guchar* end);

static gboolean save_image(const char* filename,
                           gint32 imageID, gint32 drawableID);
static gboolean write_layer(guint32 imageID, guint32 layerID,
                            FILE* stream);

static gint find_unused_index(const guchar* begin, const guchar* end,
                              gboolean alpha);

static void fill_transparent_with_index(guchar* begin, guchar* end, guchar index);

static gboolean compress_image(const GByteArray* src, GByteArray* dest);

static void write_mask(gint image_id, gint mask_id, FILE* stream);

static void write_additional_data_field(gconstpointer data, guint32 size,
                                        gchar* ID, FILE* stream);

static gboolean write_guide_as_grid(guint32 imageID, FILE* stream);


/*
 *
 * Imprimentation
 *
 */

GimpPlugInInfo PLUG_IN_INFO = {
    NULL,                       /* init_proc */
    NULL,                       /* quit_proc */
    query,                      /* query_proc */
    run,                        /* run_proc */
};

MAIN()

static
void query()
{
    static GimpParamDef load_args[] = {
        {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
        {GIMP_PDB_STRING, "filename", "The name of the file to load"},
        {GIMP_PDB_STRING, "raw_filename", "The name entered"},
    };
    static GimpParamDef load_return_vals[] = {
        {GIMP_PDB_IMAGE, "image", "Output image"},
    };
    static int nload_args = sizeof(load_args) / sizeof(load_args[0]);
    static int nload_return_vals = sizeof(load_return_vals)
        / sizeof(load_return_vals[0]);

    static GimpParamDef save_args[] = {
        {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
        {GIMP_PDB_IMAGE, "image", "Input image"},
        {GIMP_PDB_DRAWABLE, "drawable", "Drawable to save"},
        {GIMP_PDB_STRING, "filename", "The name of the file to save the image in"},
        {GIMP_PDB_STRING, "raw_filename", "The name entered"}
    };
    static int nsave_args = sizeof(save_args) / sizeof(save_args[0]);

    gimp_plugin_domain_register("gimp-dpix-plugins", LOCALEDIR);

    gimp_install_procedure("file_dpx_load",
                           "loads files of the D-Pixed DPX file format",
                           "",
                           "Masahiro SAKAI <sakai@tom.sfc.keio.ac.jp>",
                           "Masahiro SAKAI",
                           "1999-2002",
                           "<Load>/DPX",
                           NULL,
                           GIMP_PLUGIN,
                           nload_args, nload_return_vals,
                           load_args, load_return_vals);

    gimp_install_procedure("file_dpx_save",
                           "saves files in the D-Pixed DPX file format",
                           "saves files in the D-Pixed DPX file format",
                           "Masahiro SAKAI <sakai@tom.sfc.keio.ac.jp>",
                           "Masahiro SAKAI",
                           "1999-2002",
                           "<Save>/DPX",
                           "INDEXED*",
                           GIMP_PLUGIN,
                           nsave_args, 0,
                           save_args, NULL);

    gimp_register_magic_load_handler("file_dpx_load", "dpx", "", "0,string,DPX");
    gimp_register_save_handler("file_dpx_save", "dpx", "");
}



static
void run(char* name,
         int nparams,
         GimpParam* param, int* nreturn_vals, GimpParam** return_vals)
{
    static GimpParam values[2];
    gint32 run_mode = param[0].data.d_int32;
    gint32 imageID;
    gint32 drawableID;
    gint32 orig_imageID;
#ifdef GIMP_HAVE_EXPORT
    int export_result = (int)GIMP_EXPORT_CANCEL;
#endif

    *nreturn_vals = 1;
    *return_vals = values;
    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = GIMP_PDB_CALLING_ERROR;

    if (!strcmp(name, "file_dpx_load")) {
        INIT_I18N_UI();
        imageID = load_image(param[1].data.d_string);
        if (imageID == -1)
            values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
        else {
            *nreturn_vals = 2;
            values[0].data.d_status = GIMP_PDB_SUCCESS;
            values[1].type = GIMP_PDB_IMAGE;
            values[1].data.d_image = imageID;
        }
    } else if (!strcmp(name, "file_dpx_save")) {
        INIT_I18N_UI();

        imageID = orig_imageID = param[1].data.d_int32;
        drawableID = param[2].data.d_int32;

        switch ((GimpRunModeType)run_mode) {
        case GIMP_RUN_INTERACTIVE:
        case GIMP_RUN_WITH_LAST_VALS:
#ifdef GIMP_HAVE_EXPORT
            gimp_ui_init("dpx", FALSE);
            export_result =
                gimp_export_image(
                    &imageID, &drawableID, "DPX",
                    GIMP_EXPORT_CAN_HANDLE_INDEXED
                    | GIMP_EXPORT_CAN_HANDLE_ALPHA
                    | GIMP_EXPORT_CAN_HANDLE_LAYERS
                    /*| GIMP_EXPORT_CAN_HANDLE_LAYERS_AS_ANIMATION*/
                    );
            switch ((GimpExportCapabilities)export_result) {
            case GIMP_EXPORT_EXPORT:
            case GIMP_EXPORT_IGNORE:
                break;
            case GIMP_EXPORT_CANCEL:
                *nreturn_vals = 1;
                values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
                return;
                break;
            default:
                break;
            }
#endif
            break;
        default:
            break;
        }

        if (save_image(param[3].data.d_string, imageID, drawableID))
            values[0].data.d_status = GIMP_PDB_SUCCESS;
        else
            values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;

        if (export_result == GIMP_EXPORT_EXPORT)
            gimp_image_delete(imageID);
    }
}



/*
 *
 * Utility
 *
 */

static inline
guchar get_nth_bit(const guchar* data, guint index)
{
    return (data[index / 8] & BIT(index % 8)) ? 1 : 0;
}

static inline
void set_nth_bit(guchar* data, guint index, guchar value)
{
    guint byte, bit;
    byte = index / 8;
    bit  = index % 8;
    data[byte] &= ~BIT(bit);
    data[byte] |= value << bit;
}

static inline
guchar get_nth_4bit(const guchar* data, guint index)
{
    if (index % 2) {
        return (data[index / 2] >> 4) & 0x0F;
    } else {
        return data[index / 2] & 0x0F;
    }
}

static inline
void set_nth_4bit(guchar* data, guint index, guchar value)
{
    if (index % 2) {
        data[index / 2] &= 0x0F;
        data[index / 2] |= value << 4;
    } else {
        data[index / 2] &= 0xF0;
        data[index / 2] |= value;
    }
}

#ifdef HAVE_ICONV
/* these code are borrowed from mgedit-0.0.19. */

#define MGEDIT_ICONV_EILSEQ 1
#define MGEDIT_ICONV_E2BIG  2
#define MGEDIT_ICONV_EINVAL 3

#define DEBUG_ICONV(x)

void *
mgedit_iconv_open(const char *tocode, const char *fromcode)
{
	iconv_t cd, *retval;
	cd = iconv_open(tocode, fromcode);
	if (cd == (iconv_t)-1)
		return NULL;
	retval = g_new(iconv_t, 1);
	*retval = cd;
	DEBUG_ICONV(printf("mgedit_iconv_open from %s to %s\n",
			   fromcode, tocode));
	return retval;
}

void
mgedit_iconv_close(void *cdp)
{
	gint r;
	if (cdp == NULL)
		return;
	r = iconv_close(*(iconv_t *)cdp);
	assert(r == 0);
	g_free(cdp);
}

size_t
mgedit_iconv
(void *cdp, const char **inbuf, size_t *inleft, char **outbuf,
 size_t *outleft, gint *error_code)
{
	size_t r;
	DEBUG_ICONV(printf("IN  from, fromlen, to, tolen = %p %d %p %d\n",
			   inbuf ? (*inbuf) : 0,
			   inleft ? (*inleft) : 0,
			   outbuf ? (*outbuf) : 0,
			   outleft ? (*outleft) : 0));
	r = iconv(*(iconv_t *)cdp, inbuf, inleft, outbuf, outleft);
	DEBUG_ICONV(printf("OUT from, fromlen, to, tolen = %p %d %p %d\n",
			   inbuf ? (*inbuf) : 0,
			   inleft ? (*inleft) : 0,
			   outbuf ? (*outbuf) : 0,
			   outleft ? (*outleft) : 0));
	if (r == (size_t)-1) {
		DEBUG_ICONV(perror("mgedit_iconv"));
		switch (errno) {
		case EILSEQ:
			*error_code = MGEDIT_ICONV_EILSEQ;
			break;
		case E2BIG:
			*error_code = MGEDIT_ICONV_E2BIG;
			break;
		case EINVAL:
			*error_code = MGEDIT_ICONV_EINVAL;
			break;
		default:
			perror("iconv");
			abort();
		}
	}
	return r;
}

static gint
apply_iconv
(void *cdp, const gchar *buffer, gint len, gchar **buffer_r, gint *len_r,
 size_t *iconv_error_pos_r)
{
	gchar *to;
	const gchar *from_p, *from;
	size_t from_len, to_len, to_alloc;
	gint error_code;

	from_p = from = buffer;
	from_len = len;
	to_alloc = 4096;
	to = g_new(gchar, to_alloc);
	to_len = 0;
	do {
		gchar *to_p;
		size_t s, from_left, to_left;

		if (to_alloc < to_len + 4096) {
			to_alloc += 4096;
			to = (gchar*)g_realloc(to, to_alloc);
		}
		from_left = MIN(from + from_len - from_p, 256);
		to_p = to + to_len;
		to_left = 4096;
		error_code = 0;
		s = mgedit_iconv(cdp, &from_p, &from_left,
				 &to_p, &to_left, &error_code);
		if ((error_code == 0) && (from + from_len == from_p)) {
			/* Write the reset sequence. We don't realloc
			 * 'to' here in order not to hit a bug(?) in
			 * glibc-2.1. This iconv() call may cause
			 * E2BIG in theory. */
			s = mgedit_iconv(cdp, NULL, NULL,
					 &to_p, &to_left, &error_code);
			to_len = to_p - to;
			break;
		}
		to_len = to_p - to;
	} while (error_code != MGEDIT_ICONV_EILSEQ &&
		 from_p < from + from_len);
	if (error_code)
		*iconv_error_pos_r = from_p - from;
	*len_r = to_len;
	*buffer_r = to;
	return error_code;
}

#endif /* HAVE_ICONV */ /* borrowed from mgedit */


#ifdef HAVE_ICONV

static
gchar* guess_internal_codeset()
{
#ifdef GDK_USE_UTF8_MBS
    return "UTF-8";
#else
    const gchar* cp;
    gchar* codeset = NULL;
    cp = gtk_set_locale();

    /* Pick up the codeset portion from the locale name. (borrowed
     * from glibc-2.1) */
    cp += strcspn(cp, "@.+,");
    if (*cp == '.') {
        const gchar *endp = ++cp;
        while ((*endp != '\0') && (*endp != '@'))
            endp++;
        if (endp != cp) {
            gint i;
            codeset = g_strndup(cp, endp-cp);
            for (i = 0; i < endp-cp; i++)
                codeset[i] = toupper(codeset[i]);
        }
    }

    /* sigh... */
    if (codeset && !strcmp(codeset, "EUC")) {
        const struct {
            char* loc_name;
            char* codeset;
        } db[] = {
            {"ja",    "EUC-JP"},
            {"ko",    "EUC-KR"},
            {"zh_TW", "EUC-TW"},
            {"zh_CN", "EUC-CN"},
        };
        int i;

        cp = gtk_set_locale();

        for (i = 0; i < sizeof(db) / sizeof(db[0]); i++) {
            if (!strncasecmp(cp, db[i].loc_name, strlen(db[i].loc_name))) {
                g_free(codeset);
                codeset = g_strdup(db[i].codeset);
                break;
            }
        }
    }

    return codeset;
#endif
}

static
gchar* guess_external_codeset()
{
    /* FIXME: improve this list! */
    static const struct {
        char* key;
        char* codeset;
    } codeset_db[] = {
//        {"af", ""}, // Afar
//        {"am", ""}, // Abkhazian
        {"ar", "CP1256"},    // Arabic
//        {"as", ""}, // Assamese
//        {"ay", ""}, // Aymara
//        {"az", ""}, // Azerbaijani

//        {"ba", ""}, // Bashkir
//        {"be", ""}, // Byelorussian
        {"bg", "CP1251"}, // Bulgarian
//        {"bh", ""}, // Bihari
//        {"bi", ""}, // Bislama
//        {"bn", ""}, // Bengali; Bangla
//        {"bo", ""}, // Tibetan
//        {"br", ""}, // Breton

//        {"ca", ""}, // Catalan
//        {"co", ""}, // Corsican
        {"cs", "CP1250"}, // Czech
        {"cy", "CP1252"}, // Welsh

        {"da", "CP1252"}, // Danish
        {"de", "CP1252"}, // German
//        {"dz", ""}, // Bhutani

        {"el", "CP1253"}, // Greek
        {"en", "CP1252"}, // English
        {"eo", "CP1252"}, // Esperanto
        {"es", "CP1252"}, // Spanish
//        {"et", ""}, // Estonian
//        {"eu", "CP1252"}, // Basque

//        {"fa", ""}, // Persian
        {"fi", "CP1252"}, // Finnish
//        {"fj", ""}, // Fiji
//        {"fo", ""}, // Faroese
        {"fr", "CP1252"}, // French
//        {"fy", ""}, // Frisian

//        {"ga", ""}, // Irish
//        {"gd", ""}, // Scots Gaelic
//        {"gl", ""}, // Galician
//        {"gn", ""}, // Guarani
//        {"gu", ""}, // Gujarati

//        {"ha", ""}, // Hausa
        {"he", "CP1255"}, // Hebrew (formerly iw)
//        {"hi", ""}, // Hindi
//        {"hr", ""}, // Croatian
        {"hu", "CP1250"}, // Hungarian
//        {"hy", ""}, // Armenian

//        {"ia", ""}, // Interlingua
//        {"id", ""}, // Indonesian (formerly in)
//        {"ie", ""}, // Interlingue
//        {"ik", ""}, // Inupiak
//        {"is", "CP861"}, // Icelandic
        {"it", "CP1252"}, // Italian
//        {"iu", ""}, // Inuktitut

        {"ja", "SHIFT_JIS"}, /* Japanese (CP932) */
//        {"jw", ""}, // Javanese

//        {"ka", ""}, // Georgian
//        {"kk", ""}, // Kazakh
//        {"kl", ""}, // Greenlandic
//        {"km", ""}, // Cambodian
//        {"kn", ""}, // Kannada
        {"ko", "UHC"},    /* Korean (CP949) */
//        {"ko", "CP1361"}, /* Korean (Johab) ??? */
//        {"ks", ""}, // Kashmiri
//        {"ku", ""}, // Kurdish
//        {"ky", ""}, // Kirghiz

//        {"la", "CP1252"}, // Latin
//        {"ln", ""}, // Lingala
//        {"lo", ""}, // Laothian
//        {"lt", ""}, // Lithuanian
//        {"lv", ""}, // Latvian, Lettish

//        {"mg", ""}, // Malagasy
//        {"mi", ""}, // Maori
//        {"mk", ""}, // Macedonian
//        {"ml", ""}, // Malayalam
//        {"mn", ""}, // Mongolian
//        {"mo", ""}, // Moldavian
//        {"mr", ""}, // Marathi
//        {"ms", ""}, // Malay
//        {"mt", ""}, // Maltese
//        {"my", ""}, // Burmese

//        {"na", ""}, // Nauru
//        {"ne", ""}, // Nepali
        {"nl", "CP1252"}, // Dutch
//        {"no", "CP1252"}, // Norwegian

//        {"oc", ""}, // Occitan
//        {"om", ""}, // (Afan) Oromo
//        {"or", ""}, // Oriya

//        {"pa", ""}, // Punjabi
//        {"pl", ""}, // Polish
//        {"ps", ""}, // Pashto, Pushto
//        {"pt", "CP860"}, // Portuguese

//        {"qu", ""}, // Quechua

//        {"rm", ""}, // Rhaeto-Romance
//        {"rn", ""}, // Kirundi
//        {"ro", ""}, // Romanian
        {"ru", "CP1251"}, // Russian
//        {"rw", ""}, // Kinyarwanda

//        {"sa", ""}, // Sanskrit
//        {"sd", ""}, // Sindhi
//        {"sg", ""}, // Sangho
//        {"sh", ""}, // Serbo-Croatian
//        {"si", ""}, // Sinhalese
//        {"sk", ""}, // Slovak
//        {"sl", ""}, // Slovenian
//        {"sm", ""}, // Samoan
//        {"sn", ""}, // Shona
//        {"so", ""}, // Somali
//        {"sq", ""}, // Albanian
//        {"sr", ""}, // Serbian
//        {"ss", ""}, // Siswati
//        {"st", ""}, // Sesotho
//        {"su", ""}, // Sundanese
        {"sv", "CP1252"}, // Swedish
//        {"sw", ""}, // Swahili

//        {"ta", ""}, // Tamil
//        {"te", ""}, // Telugu
//        {"tg", ""}, // Tajik
        {"th", "CP874"}, // Thai
//        {"ti", ""}, // Tigrinya
//        {"tk", ""}, // Turkmen
//        {"tl", ""}, // Tagalog
//        {"tn", ""}, // Setswana
//        {"to", ""}, // Tonga
        {"tr", "CP1254"}, // Turkish
//        {"ts", ""}, // Tsonga
//        {"tt", ""}, // Tatar
//        {"tw", ""}, // Twi

//        {"ug", ""}, // Uighur
//        {"uk", ""}, // Ukrainian
//        {"ur", ""}, // Urdu
//        {"uz", ""}, // Uzbek

        {"vi", "CP1258"}, // Vietnamese
//        {"vo", ""}, // Volapuk

//        {"wo", ""}, // Wolof

//        {"xh", ""}, // Xhosa

//        {"yi", ""}, // Yiddish (formerly ji)
//        {"yo", ""}, // Yoruba

//        {"za", ""}, // Zhuang
        {"zh_cn", "GBK"},  /* Simplified Chinese (CP936) */
        {"zh_tw", "BIG5"}, /* Traditional Chinese (CP950) */
        {"zh",    "BIG5"}, /* Chinese (default???) */
//        {"zu", ""}, // Zulu
    };

    char* locale;
    char* result;
    guint i;

    locale = g_strdup(gtk_set_locale());

    for (i = 0; i < strlen(locale); i++)
        locale[i] = tolower(locale[i]);

    result = NULL;

    for (i = 0; i < sizeof(codeset_db) / sizeof(codeset_db[0]); i++) {
        if (!strncmp(locale, codeset_db[i].key, strlen(codeset_db[i].key))) {
            result = codeset_db[i].codeset;
            break;
        }
    }

    g_free(locale);

    return result;
}

#endif /* HAVE_ICONV */


static
gchar* convert_to_internal(const char* src)
{
#if defined(_WIN32) && defined(GDK_USE_UTF8_MBS)
    return g_filename_to_utf8(src);
#elif defined(HAVE_ICONV)
    void* cdp;
    gchar* dest = NULL;
    gint dest_len;
    size_t error_code;
    gchar* internal_code;
    gchar* external_code;

    internal_code = guess_internal_codeset();
    external_code = guess_external_codeset();
    if (!internal_code || !external_code)
        return NULL;
    cdp = mgedit_iconv_open(internal_code, external_code);
    g_free(internal_code);
    if (!cdp)
        return NULL;
    apply_iconv(cdp, src, strlen(src) + 1, &dest, &dest_len, &error_code);
    mgedit_iconv_close(cdp);
    return dest;
#else
    return NULL;
#endif
}

static
gchar* convert_to_external(const char* src)
{
#if defined(_WIN32) && defined(GDK_USE_UTF8_MBS)
    return g_filename_from_utf8(src);
#elif defined(HAVE_ICONV)
    void* cdp;
    gchar* dest = NULL;
    gint dest_len;
    size_t error_code;
    gchar* internal_code;
    gchar* external_code;

    internal_code = guess_internal_codeset();
    external_code = guess_external_codeset();
    if (!internal_code || !external_code)
        return NULL;
    cdp = mgedit_iconv_open(external_code, internal_code);
    g_free(internal_code);
    if (!cdp)
        return NULL;
    apply_iconv(cdp, src, strlen(src) + 1, &dest, &dest_len, &error_code);
    mgedit_iconv_close(cdp);
    return dest;
#else
    return NULL;
#endif
}


/*
 *
 * Load
 *
 */

gint32 load_image(const char* filename)
{
    FILE* stream;
    DPIXED_FILEHEADER fileheader;
    gint imageID;
    guint buffer_size;
    guchar* buffer;
    guchar* buffer_end;
    guchar* current_pos;
    guchar cmap[256 * 3];
    char* progress;
    guint i;
#ifdef G_OS_WIN32
    guint64 filesize;
#else
    off_t filesize;
#endif

    /* Progress Display Initialization  */
    if (strrchr(filename, G_DIR_SEPARATOR) != NULL)
        progress = g_strdup_printf(_("Loading %s:"), strrchr(filename, G_DIR_SEPARATOR) + 1);
    else
        progress = g_strdup_printf(_("Loading %s:"), filename);
    gimp_progress_init(progress);
    g_free(progress);

    stream = fopen(filename, "rb");
    /* Is this a valid File?
       Should never be used, because gimp tests it. */
    if (!stream) {
        g_message(_("%s: can't open \"%s\""), PROG_NAME, filename);
        gimp_quit();
        return -1;
    }

    /* Getting filesize */
    {
#ifdef G_OS_WIN32
        HANDLE hFind;
        WIN32_FIND_DATA data;
        hFind = FindFirstFile(filename, &data);
        filesize = ((guint64)data.nFileSizeHigh << 32) | (guint64)data.nFileSizeLow;
        FindClose(hFind);
#else
        struct stat file_stat;
        if (stat(filename, &file_stat))
            gimp_quit();
        filesize = file_stat.st_size;
#endif
    }

    /* File Header */
    fread(&fileheader, sizeof(DPIXED_FILEHEADER), 1, stream);
    fileheader.version = GUINT32_FROM_LE(fileheader.version);
    fileheader.dwFlag  = GUINT32_FROM_LE(fileheader.dwFlag);
    fileheader.width   = GUINT32_FROM_LE(fileheader.width);
    fileheader.height  = GUINT32_FROM_LE(fileheader.height);
    fileheader.nBitmap = GUINT32_FROM_LE(fileheader.nBitmap);

    /* Is it a DPX file? */
    if (strncmp(fileheader.fileDescription, "DPX", 3)) {
        g_message(_("%s: %s is not a valid DPX file.\n"), PROG_NAME, filename);
        return -1;
    }

    imageID = gimp_image_new(fileheader.width, fileheader.height, GIMP_INDEXED);
    gimp_image_set_filename(imageID, (char*)filename);

    /* Color Map */
    if (fileheader.dwFlag & DPF_FILE_HASGLOBALCOLORTABLE) {
        for (i = 0; i < 256; i++) {
            DP_RGBQuad rgb;
            fread(&rgb, sizeof(rgb), 1, stream);
            cmap[i * 3 + 0] = rgb.rgbRed;
            cmap[i * 3 + 1] = rgb.rgbGreen;
            cmap[i * 3 + 2] = rgb.rgbBlue;
        }
        gimp_image_set_cmap(imageID, cmap, 256);
    } else {
        g_message(_("%s: %s has no global color table."), PROG_NAME, filename);
    }

    /* Allocate Working Memory */
    buffer_size = filesize - ftell(stream);
    buffer = (guchar*)g_malloc(buffer_size);
    fread(buffer, buffer_size, 1, stream);
    fclose(stream);
    current_pos = buffer;
    buffer_end  = buffer + buffer_size;

    /* Layers */
    gimp_progress_update(0);
    for (i = 0; i < fileheader.nBitmap; i++) {
        current_pos += read_layer(current_pos, buffer_end, imageID);
        gimp_progress_update((gdouble)(i + 1) / fileheader.nBitmap);
    }

    if (fileheader.nBitmap == 0)
        g_message(_("%s: image has no layer."), PROG_NAME);

    /* Aditional Data Field  */
    current_pos += read_additional_data_field(imageID, current_pos, buffer_end);

    g_free(buffer);

    return imageID;
}


static
guint32 read_layer(const guchar* buffer, const guchar* buffer_end, guint32 imageID)
{
    DPIXED_FILEIMAGEHEADER header;
    gboolean use_alpha;
    gchar layer_name[65];
    gchar* layer_name2;
    gint layerID;
    guint width, height;
    const guchar* current_pos;
    const guchar* colormask = NULL;

    width  = gimp_image_width(imageID);
    height = gimp_image_height(imageID);

    current_pos = buffer;

    memcpy(&header, current_pos, sizeof(header));
    header.dwSize    = GUINT32_FROM_LE(header.dwSize);
    header.dwFlag    = GUINT32_FROM_LE(header.dwFlag);
    header.keyColor  = GUINT32_FROM_LE(header.keyColor);
    header.alpha     = GUINT32_FROM_LE(header.alpha);
    header.maskBit   = GUINT32_FROM_LE(header.maskBit);
    header.bgColor   = GUINT32_FROM_LE(header.bgColor);
    header.imageSize = GUINT32_FROM_LE(header.imageSize);
    header.reserved  = GUINT32_FROM_LE(header.reserved);
    header.left             = GINT32_FROM_LE(header.left);
    header.top              = GINT32_FROM_LE(header.top);
    header.right            = GINT32_FROM_LE(header.right);
    header.bottom           = GINT32_FROM_LE(header.bottom);
    header.delay            = GINT32_FROM_LE(header.delay);
    header.animeMode        = GINT32_FROM_LE(header.animeMode);
    header.animeInput       = GINT32_FROM_LE(header.animeInput);
    header.animeTransparent = GINT32_FROM_LE(header.animeTransparent);

    if (header.dwFlag & DPF_IMAGE_COLORMASK)
        colormask = current_pos + sizeof(header);

    current_pos += header.dwSize;

    if (header.dwFlag & DPF_IMAGE_HASLOCALCOLORTABLE) {
        current_pos += sizeof(DP_RGBQuad) * 256;
        g_message(_("%s: Layer color table is not supported."), PROG_NAME);
    }

    memset(layer_name, 0, sizeof(layer_name));
    strncpy(layer_name, header.description, sizeof(header.description));
    layer_name2 = convert_to_internal(layer_name);

    if (header.keyColor >= 0 && header.keyColor < 256)
        use_alpha = TRUE;
    else
        use_alpha = FALSE;

    layerID = gimp_layer_new(
        imageID,
        layer_name2 ? layer_name2 : layer_name,
        width, height,
        use_alpha ? GIMP_INDEXEDA_IMAGE : GIMP_INDEXED_IMAGE,
        (double) header.alpha * 100 / 16,
        GIMP_NORMAL_MODE);

    g_free(layer_name2);

    gimp_layer_set_visible(layerID, header.dwFlag & DPF_IMAGE_ENABLE);
    gimp_image_add_layer(imageID, layerID, 0);

    {
        guint bmp_linesize;
        guchar* bmp;
        guchar* mask;

        bmp_linesize = ((width + 3) / 4) * 4;

        if (header.dwFlag & DPF_IMAGE_IMAGECOMPRESSED) {
            guint bmp_size;
            bmp_size = bmp_linesize * height;
            bmp = (guchar*)g_malloc(bmp_size);
            if (!decompress_image(current_pos, header.imageSize, bmp, bmp_size)) {
                g_message("BUG: decompress_image() failed\n");
                gimp_quit();
            }
        } else {
            bmp = (guchar*)current_pos;
        }

        current_pos += header.imageSize;
        if (current_pos > buffer_end) {
            g_message(_("%s: invalid image (or you encount a bug :)\n"
                        "header.imageSize = %d\n"
                        "current_pos - buffer_end = %X\n"),
                      PROG_NAME, header.imageSize, current_pos - buffer_end);
            gimp_quit();
        }

        if (header.dwFlag & DPF_IMAGE_HASMASK) {
            guint size;
            mask = (guchar*)g_malloc(width * height);
            size = read_mask(&header, current_pos, mask, width, height);
            current_pos += size;
            if (current_pos > buffer_end) {
                g_message(_("%s: invalid file (or you encount a bug :)\n"
                            "encoded mask size = %d\n"
                            "current_pos - buffer_end = %X\n"),
                          PROG_NAME, size, current_pos - buffer_end);
                gimp_quit();
            }
        } else {
            mask = NULL;
            header.dwFlag &= ~DPF_IMAGE_USEALPHAPLANE;
        }
        
        {
            guchar *dest;
            guint x, y;
            guint pixel_size;

            if (use_alpha)
                pixel_size = 2;
            else
                pixel_size = 1;

            dest = (guchar*)g_malloc(width * height * pixel_size);

            /* image */
            for (y = 0; y < height; y++) {
                guchar* src_scanline;
                guchar* dst_scanline;
                src_scanline = bmp + bmp_linesize * ((height - 1) - y);
                dst_scanline = dest + width * pixel_size * y;
                for (x = 0; x < width; x++)
                    dst_scanline[x * pixel_size] = src_scanline[x];
            }

            /* mask */
            if ((header.dwFlag & DPF_IMAGE_HASMASK)
                && (header.dwFlag & DPF_IMAGE_COLORMASK)) {
                for (y = 0; y < height; y++) {
                    guchar* dst_scanline;
                    guchar* mask_scanline;
                    dst_scanline  = dest + width * pixel_size * y;
                    mask_scanline = mask + width * y;
                    for (x = 0; x < width; x++)
                        if (colormask[dst_scanline[x * pixel_size]])
                            mask_scanline[x] = 0;
                }
            }

            /* alpha */
            if (use_alpha) {
                for (y = 0; y < height; y++) {
                    guchar* dst_scanline;
#ifndef LAYERMASK_READ
                    guchar* mask_scanline;
                    mask_scanline = mask + width * y;
#endif
                    dst_scanline  = dest + width * pixel_size * y;
                    for (x = 0; x < width; x++) {
                        if (dst_scanline[x * pixel_size] == header.keyColor)
                            dst_scanline[x * pixel_size + 1] = 0;
#ifndef LAYERMASK_READ
                        else if (header.dwFlag & DPF_IMAGE_USEALPHAPLANE)
                            dst_scanline[x * pixel_size + 1] = mask_scanline[x];
#endif
                        else
                            dst_scanline[x * pixel_size + 1] = 0xFF;
                    }
                }
            }

            /* image */
            {
                GimpDrawable* drawable;
                GimpPixelRgn pixel_rgn;
                drawable = gimp_drawable_get(layerID);
                gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0,
                                    width, height, TRUE, FALSE);
                /* this statement cause crash with a certain file. why? */
                gimp_pixel_rgn_set_rect(&pixel_rgn, dest, 0, 0, width, height);
                /* */
                gimp_drawable_flush(drawable);
                gimp_drawable_detach(drawable);
            }

            /* layer mask */
#ifdef LAYERMASK_READ
            if (header.dwFlag & DPF_IMAGE_USEALPHAPLANE) {
                gint mask_id;
                GimpDrawable *drawable;
                GimpPixelRgn pixel_rgn;
                guint i;
                for (i = 0; i < width * height; i++)
                    mask[i] = ~mask[i];
                mask_id = gimp_layer_create_mask(layerID, GIMP_WHITE_MASK);
                gimp_image_add_layer_mask(imageID, layerID, mask_id);
                drawable = gimp_drawable_get(mask_id);
                gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0,
                                    width, height, TRUE, FALSE);            
                gimp_pixel_rgn_set_rect(&pixel_rgn, mask, 0, 0, width, height);
                gimp_drawable_flush(drawable);
                gimp_drawable_detach(drawable);
            }
#endif

            g_free(dest);
        }

        if (header.dwFlag & DPF_IMAGE_HASMASK)
            g_free(mask);
        
        if (header.dwFlag & DPF_IMAGE_IMAGECOMPRESSED)
            g_free(bmp);
    }

    return current_pos - buffer;
}


static
gboolean decompress_image(const guchar* src_data, guint src_size,
                          guchar* dest_data, guint dest_size)
{
    const guchar* src_buf = src_data;
    const guchar* src_buf_end = src_buf + src_size;
    guchar* dst_buf = dest_data;
    guchar* dst_buf_end = dst_buf + dest_size;

    while (dst_buf < dst_buf_end && src_buf < src_buf_end) {
        if (src_buf + 1 < src_buf_end && src_buf[0] == src_buf[1]) {
            if (dst_buf + src_buf[2] <= dst_buf_end) {
                memset(dst_buf, src_buf[0], src_buf[2]);
                dst_buf += src_buf[2];
                src_buf += 3;
            } else {
                g_message(_("%s: a broken DPX file.\n"), PROG_NAME);
                return FALSE;
            }
        } else {
            *(dst_buf++) = *(src_buf++);
        }
    }
    return TRUE;
}


static
guint read_mask(const DPIXED_FILEIMAGEHEADER* header,
                const guchar* src, guchar* dst,
                guint width, guint height)
{
    guint bitcount;
    guint result;

    /* bitcount */
    if (header->dwFlag & DPF_IMAGE_MASKCOMPRESSED)
        bitcount = 1;
    else {
        switch (header->maskBit) {
        case 1:
        case 4:
        case 8:
            bitcount = header->maskBit;
            break;
        default:
            g_error(_("%s: Mask type is unknown."), PROG_NAME);
            return FALSE;
        }
    }

    if (header->dwFlag & DPF_IMAGE_MASKCOMPRESSED) {
        guchar* tmp;
        guint mask_size_4bit, mask_size;        

        mask_size_4bit = ((width + 3) / 4) * height;
        mask_size = (mask_size_4bit + 1) / 2;

        tmp = (guchar*)g_malloc(mask_size);
        result = decompress_mask(src, tmp, mask_size_4bit);
        {
            guint x, y;
            guint line_bit = ((width + 3) / 4) * 4;
            for (y = 0; y < height; y++) {
                guint line_start_bit;
                guchar* dest_line;
                line_start_bit = ((height - 1) - y) * line_bit;
                dest_line = dst + y * width;
                for (x = 0; x < width; x++) {
                    if(get_nth_bit(tmp, line_start_bit + x))
                        dest_line[x] = 0xFF;
                    else
                        dest_line[x] = 0;
                }
            }
        }
        g_free(tmp);
    } else {
        guint linesize, mask_size;
        guint x, y;
        g_message(_("%s: Normal(not compressed) mask feature isn't tested yet. ;-)"), PROG_NAME);
        /* Each start of line is aligned to DWORD. */
        linesize = ((bitcount * width + 31) / 32) * 4;
        mask_size = linesize * height;
        for (y = 0; y < height; y++) {
            const guchar* src_line = src + linesize * ((height - 1) - y);
            guchar* dst_line = dst + width * y;;
            switch (bitcount) {
            case 1:
                for (x = 0; x < width; x++)
                    dst_line[x] = get_nth_bit(src_line, x) ? 0xFF : 0;
                break;
            case 4:
                for (x = 0; x < width; x++) {
                    guchar tmp = get_nth_4bit(src_line, x);
                    dst_line[x] = (tmp << 4) | tmp;
                }
                break;
            case 8:
                memcpy(dst_line, src_line, width);
                break;
            }
        }
        result = mask_size;
    }

    return result;
}    


static
guint decompress_mask(const guchar* src, guchar* dst, guint decompressed_size_4bit)
{
    guint pos;
    guint dest_pos;
    guint size_4bit;
    guint i;

    pos = 0;
    dest_pos = 0;
    size_4bit = decompressed_size_4bit;

    while (dest_pos < size_4bit) {
        guchar v1, v2;
        v1 = get_nth_4bit(src, pos);
        if (dest_pos + 1 == size_4bit) {
            set_nth_4bit(dst, pos, v1);
            pos++;
            break;
        }
        v2 = get_nth_4bit(src, pos + 1);
        if (v1 == v2) {
            guint len;
            len = get_nth_4bit(src, pos + 2) 
                | (get_nth_4bit(src, pos + 3) << 4);
            if (dest_pos + len > size_4bit) {
                g_message(_("%s: decompress_mask: Invalid mask."), PROG_NAME);
                break;
            }
            for (i = 0; i < len; i++)
                set_nth_4bit(dst, dest_pos + i, v1);
            pos += 4;
            dest_pos += len;
        } else {
            set_nth_4bit(dst, dest_pos, v1);
            pos++;
            dest_pos++;
        }
    }

    return (pos + 1) / 2;
}


#define ADF_ID_SIZE 4

static
guint read_additional_data_field(gint imageID,
                                 const guchar* begin, const guchar* end)
{
    const guchar* current_pos = begin;

    while (current_pos + 8 <= end) {
        typedef struct {
            char* name;
            void (*reader)(gint imageID,
                           const guchar* begin, const guchar* end);
        } ADF_ITEM;

        const ADF_ITEM list[] = {
            {"GRID", read_grid},
            {"FRAM", read_frame},
            {"SHAP", read_shape},
            {"EOF ", NULL},
        };

        const ADF_ITEM* item = NULL;
        guint32 length;
        guint i;

        for (i = 0; i < sizeof(list) / sizeof(list[0]); i++) {
            if (!strncmp((char*)current_pos, list[i].name, ADF_ID_SIZE)) {
                item = &list[i];
                break;
            }
        }
        current_pos += ADF_ID_SIZE;

        length = GUINT32_FROM_LE(*(guint32*)current_pos);
        current_pos += sizeof(guint32);

        if (item && item->reader)
            item->reader(imageID, current_pos, current_pos + length);
        current_pos += length;
    }

    return current_pos - begin;
}


static
void read_grid(gint imageID, const guchar* begin, const guchar* end)
{
    DPGRID_HEADER header;
    guint i;
    guint16* guides;

    if (begin + sizeof(DPGRID_HEADER) > end) {
        g_message(_("%s: read_grid: Invalid file, data size is too small.\n"
                    "Grid data field is ignored.\n"
                    "field size = %d, header size = %d"),
                  PROG_NAME, end - begin, sizeof(DPGRID_HEADER));
        return;
    }

    if (strncmp((const char*)begin, "DGR", 3)) {
        g_message(_("%s: read_grid: Invalid grid format.\n"), PROG_NAME);
        return;
    }

    memcpy(&header, begin, sizeof(header));
    header.m_Width    = GUINT32_FROM_LE(header.m_Width);
    header.m_Height   = GUINT32_FROM_LE(header.m_Height);
    header.m_XGrid    = GUINT32_FROM_LE(header.m_XGrid);
    header.m_YGrid    = GUINT32_FROM_LE(header.m_YGrid);
    header.m_Flags    = GUINT32_FROM_LE(header.m_Flags);
    header.m_Reserved = GUINT32_FROM_LE(header.m_Reserved);

    guides = (guint16*)(begin + sizeof(header));

    if ((guchar*)(guides + header.m_XGrid + header.m_YGrid) > end) {
        g_message(_("%s: read_grid: broken file!\n"
                    "data size is too small. some grids will be ignored.\n"
                    "field size = %d, header size = %d\n"
                    "header.m_XGrid = %d,  header.m_YGrid = %d\n"),
                  PROG_NAME, end - begin, sizeof(DPGRID_HEADER),
                  header.m_XGrid, header.m_YGrid);

        {
            guint grid_count;
            grid_count = (end - (guchar*)guides) / 2;
            
            if (header.m_XGrid < grid_count) {
                header.m_YGrid = grid_count - header.m_XGrid;
            } else {
                header.m_XGrid = grid_count;
                header.m_YGrid = 0;
            }
        }
    }

    for (i = 0; i < header.m_XGrid; i++)
        gimp_image_add_vguide(imageID, GUINT16_FROM_LE(guides[i]));
    for (i = header.m_XGrid; i < header.m_XGrid + header.m_YGrid; i++)
        gimp_image_add_hguide(imageID, GUINT16_FROM_LE(guides[i]));
}


static
void read_frame(gint imageID, const guchar* begin, const guchar* end)
{
#ifdef GIMP_HAVE_PARASITES
    gimp_image_attach_new_parasite(imageID, "dpx/frame",
                                   0/*GIMP_PARASITE_PERSISTENT*/,
                                   end - begin,
                                   (const gpointer)begin);
#endif
}


static
void read_shape(gint imageID, const guchar* begin, const guchar* end)
{
#ifdef GIMP_HAVE_PARASITES
    gimp_image_attach_new_parasite(imageID, "dpx/shape",
                                   0/*GIMP_PARASITE_PERSISTENT*/,
                                   end - begin,
                                   (const gpointer)begin);
#endif
}




/*
 *
 * Save
 *
 */

static
gboolean save_image(const char* filename, gint32 imageID, gint32 drawableID)
{
    FILE* stream;
    DPIXED_FILEHEADER header;
    gchar* progress;
    GimpImageBaseType img_type;
    gint32* layers;
    gint nlayers;
    gint ncolors;
    guchar* cmap;
    int i;
    gboolean has_guide;
#ifdef GIMP_HAVE_PARASITES
    GimpParasite* frame;
    GimpParasite* shape;
#endif

    /* Is valid data? */
    img_type = gimp_image_base_type(imageID);
    if (img_type != GIMP_INDEXED)
        return FALSE;

    /* Progress Display Initialization  */
    if (strrchr(filename, G_DIR_SEPARATOR) != NULL)
        progress = g_strdup_printf(_("Saving %s:"), strrchr(filename, G_DIR_SEPARATOR) + 1);
    else
        progress = g_strdup_printf(_("Saving %s:"), filename);
    gimp_progress_init(progress);
    g_free(progress);

    /* open the destination file for writing */
    stream = fopen(filename, "wb");
    if (!stream) {
        g_message(_("can't open %s\n"), filename);
        return FALSE;
    }

    /* Is there any guides? */
    if (gimp_image_find_next_guide(imageID, 0) != -1)
        has_guide = TRUE;
    else
        has_guide = FALSE;

#ifdef GIMP_HAVE_PARASITES
    frame = gimp_image_parasite_find (imageID, "dpx/frame");
    shape = gimp_image_parasite_find (imageID, "dpx/shape");
#endif

    /* get layers */
    layers = gimp_image_get_layers(imageID, &nlayers);

    /* write file header */
    memset(&header, 0, sizeof(header));
    sprintf(header.fileDescription,
            "DPX - D-Pixed image file.\n"
            "Written by GIMP/%d.%d.%d with DPX-plugin/%s\n\n",
            gimp_major_version, gimp_minor_version, gimp_micro_version,
            VERSION);
    header.version = GUINT32_TO_LE(0x2110);
    header.dwFlag  = DPF_FILE_HASGLOBALCOLORTABLE;
    if (has_guide)
        header.dwFlag |= DPF_FILE_HASGRID;
#ifdef GIMP_HAVE_PARASITES
    if (frame)
        header.dwFlag |= DPF_FILE_HASFRAME;
    if (shape)
        header.dwFlag |= DPF_FILE_HASSHAPE;
#endif
    header.dwFlag  = GUINT32_TO_LE(header.dwFlag);
    header.width   = GUINT32_TO_LE(gimp_image_width(imageID));
    header.height  = GUINT32_TO_LE(gimp_image_height(imageID));
    header.nBitmap = GUINT32_TO_LE(nlayers);
    fwrite(&header, sizeof(header), 1, stream);

    /* write color table  */
    cmap = gimp_image_get_cmap(imageID, &ncolors);
    for (i = 0; i < MIN(ncolors, 256); i++) {
        DP_RGBQuad rgb;
        rgb.rgbRed   = cmap[i * 3 + 0];
        rgb.rgbGreen = cmap[i * 3 + 1];
        rgb.rgbBlue  = cmap[i * 3 + 2];
        rgb.rgbReserved = 0;
        fwrite(&rgb, sizeof(rgb), 1, stream);
    }
    if (ncolors < 256) {
        DP_RGBQuad rgb;
        rgb.rgbRed   = 0;
        rgb.rgbGreen = 0;
        rgb.rgbBlue  = 0;
        rgb.rgbReserved = 0;
        for (i = ncolors; i < 256; i++)
            fwrite(&rgb, sizeof(rgb), 1, stream);
    }

    /* Layers */
    gimp_progress_update(0);
    for (i = 0; i < nlayers; i++) {
        if (!write_layer(imageID, layers[nlayers - 1 - i], stream))
            return FALSE;
        gimp_progress_update((gdouble)(i + 1) / nlayers);
    }

    /* Additional Data Field */

    /* Guide => Grid */
    if (has_guide)
        write_guide_as_grid(imageID, stream);

#ifdef GIMP_HAVE_PARASITES
    /* FRAME */
    if (frame)
        write_additional_data_field(gimp_parasite_data(frame),
                                    gimp_parasite_data_size(frame),
                                    "FRAM", stream);

    /* SHAPE */
    if (shape)
        write_additional_data_field(gimp_parasite_data(shape),
                                    gimp_parasite_data_size(shape),
                                    "SHAP", stream);
#endif

    /* End Of File */
    write_additional_data_field(NULL, 0, "EOF ", stream);

    fclose(stream);
    gimp_progress_update(1);
    return TRUE;
}


static
gboolean write_layer(guint32 imageID, guint32 layerID, FILE* stream)
{
    GByteArray* win_bmp;
    gint width, height;
    gint cols, rows;
    gint offset_x, offset_y;
    gint win_bmp_linesize;
    gboolean has_alpha;
    gboolean alpha_as_mask = FALSE;
    gint unused_index;
    gint32 mask_id;

    switch (gimp_drawable_type(layerID)) {
    case GIMP_INDEXEDA_IMAGE:
        has_alpha  = TRUE;
        break;
    case GIMP_INDEXED_IMAGE:
        has_alpha  = FALSE;
        break;
    default:
        g_message(_("%s: Unexpected DrawableType."), PROG_NAME);
        return FALSE;
    }

    width  = gimp_image_width(imageID);
    height = gimp_image_height(imageID);

    {
        GimpDrawable *drawable;
        GimpPixelRgn pixel_rgn;
        guchar *buf;
        guint bpp;
        guint buf_size;
        int x, y;

        drawable = gimp_drawable_get(layerID);
        gimp_drawable_offsets(layerID, &offset_x, &offset_y);
        bpp = drawable->bpp;

        cols = MIN(offset_x + (gint)drawable->width, width) - offset_x;
        rows = MIN(offset_y + (gint)drawable->height, height) - offset_y;

        buf_size = cols * rows * bpp;
        buf = (guchar*)g_malloc(buf_size);

        gimp_pixel_rgn_init(&pixel_rgn, drawable, offset_x, offset_y, cols, rows, FALSE, FALSE);
        gimp_pixel_rgn_get_rect(&pixel_rgn, buf, offset_x, offset_y, cols, rows);
        gimp_drawable_detach(drawable);
        unused_index = find_unused_index(buf, buf + buf_size, has_alpha);
        if (unused_index == -1) {
            unused_index = 0xFF;
            if (has_alpha)
                g_message(_("there are no color to use as a transparent color\n"));
        }
        if (has_alpha)
            fill_transparent_with_index(buf, buf + buf_size, unused_index);

        win_bmp_linesize = ((width + 3) / 4) * 4;
        win_bmp = g_byte_array_new();
        g_byte_array_set_size(win_bmp, win_bmp_linesize * height);
        memset(win_bmp->data, unused_index, win_bmp->len);

        for (y = 0; y < rows; y++) {
            guchar* src;
            guchar* dest;
            src  = buf + cols * y * bpp;
            dest = win_bmp->data + (height - 1 - (y + offset_y)) * win_bmp_linesize + offset_x;
            for (x = 0; x < cols; x++) {
                *dest = src[0];
//                if (alpha_as_mask)
//                    *mask = src[bpp - 1];
                src += bpp;
                dest++;
            }
        }

        g_free(buf);
    }

#ifdef LAYERMASK_WRITE
    mask_id = gimp_layer_get_mask_id(layerID);
#else
    mask_id = -1;
#endif

    {
        GByteArray *compressed_image;
        GByteArray *output_image;
        DPIXED_FILEIMAGEHEADER header;
        gboolean is_image_compressed;

        compressed_image = g_byte_array_new();
        is_image_compressed = compress_image(win_bmp, compressed_image);
        if (is_image_compressed) {
            output_image = compressed_image;
            g_byte_array_free(win_bmp, TRUE);
        } else {
            output_image = win_bmp;
            g_byte_array_free(compressed_image, TRUE);
        }

        memset(&header, 0, sizeof(header));

        {
            gchar* src;
            gchar* dst;
            src = gimp_layer_get_name(layerID);
            dst = convert_to_external(src);
            strncpy(header.description, dst ? dst : src,
                    sizeof(header.description));
        }

        if (mask_id != -1)
            header.dwFlag |= DPF_IMAGE_HASMASK
                | DPF_IMAGE_MASKCOMPRESSED | DPF_IMAGE_USEALPHAPLANE;
        if (0)
            header.dwFlag |= DPF_IMAGE_HASLOCALCOLORTABLE;
        if (is_image_compressed)
            header.dwFlag |= DPF_IMAGE_IMAGECOMPRESSED;
        if (gimp_layer_get_visible(layerID))
            header.dwFlag |= DPF_IMAGE_ENABLE;
        if (0)
            header.dwFlag |= DPF_IMAGE_COLORMASK;

        header.keyColor  = unused_index;

        header.dwSize    = GUINT32_TO_LE(sizeof(header));
        header.dwFlag    = GUINT32_TO_LE(header.dwFlag);
        header.keyColor  = GUINT32_TO_LE(header.keyColor);
        header.alpha     = GUINT32_TO_LE(gimp_layer_get_opacity(layerID) * 16 / 100);
        header.maskBit   = GUINT32_TO_LE(1);
        header.bgColor   = GUINT32_TO_LE(0xFF);
        header.imageSize = GUINT32_TO_LE(output_image->len);
        header.reserved  = GUINT32_TO_LE(0);
        header.left   = GINT32_TO_LE(offset_x);
        header.top    = GINT32_TO_LE(offset_y);
        header.right  = GINT32_TO_LE(offset_x + cols - 1);
        header.bottom = GINT32_TO_LE(offset_y + rows - 1);
        header.delay  = GINT32_TO_LE(0);
        header.animeMode  = GINT32_TO_LE(0);
        header.animeInput = GINT32_TO_LE(0);
        header.animeTransparent = GINT32_TO_LE(0);
        fwrite(&header, sizeof(header), 1, stream);

        /* image */
        fwrite(output_image->data, output_image->len, 1, stream);
        g_byte_array_free(output_image, TRUE);
    }

    /* layer mask*/
    if (mask_id != -1)
        write_mask(imageID, mask_id, stream);

    return TRUE;
}


static
gint find_unused_index(const guchar* begin, const guchar* end, gboolean alpha)
{
    guchar map[256];
    const guchar* data;
    int i;

    memset(&map, 0, sizeof(map));

    if (alpha) {
        for (data = begin; data < end; data+=2) {
            if (data[1] != 0)
                map[*data] = 1;
        }
    } else {
        for (data = begin; data < end; data++)
            map[*data] = 1;
    }

    for (i = 255; i >= 0; i--) {
        if (map[i] == 0)
            return i;
    }

    return -1;
}


static
void fill_transparent_with_index(guchar* begin, guchar* end, guchar index)
{
    guchar* data;
    for (data = begin; data < end; data+=2) {
        if (data[1] < 128)
            data[0] = index;
    }
}


/* Image compresser */

static
gboolean compress_image(const GByteArray* src, GByteArray* dest)
{
    const guchar* src_pos = src->data;
    const guchar* src_end = src_pos + src->len;
    guchar* dest_pos;
    guchar* dest_end;
    guchar count, value;

    g_byte_array_set_size(dest, src->len);
    dest_pos = dest->data;
    dest_end = dest_pos + dest->len;

    while (src_pos < src_end) {
        value = *src_pos;
        src_pos++;
        count = 1;
        while ((src_pos < src_end) && (value == *src_pos) && (count < 255)) {
            count++;
            src_pos++;
        }
        if (count == 1) {
            if (dest_pos < dest_end)
                *(dest_pos++) = value;
            else
                return FALSE;
        } else {
            if (dest_pos + 2 < dest_end) {
                dest_pos[0] = dest_pos[1] = value;
                dest_pos[2] = count;
                dest_pos += 3;
            } else
                return FALSE;
        }
    }
    g_byte_array_set_size(dest, dest_pos - dest->data);
    return TRUE;
}


static
void write_4bit(guchar value, gint* state, FILE* stream)
{
    if (*state == -1)
        *state = value;
    else {
        guchar tmp;
        tmp = (value << 4) & *state;
        fwrite(&tmp, 1, 1, stream);
        *state = -1;
    }
}

static
void write_mask(gint image_id, gint mask_id, FILE* stream)
{
    GimpDrawable* drawable;
    GimpPixelRgn pixel_rgn;
    gint width, height;
    gint cols, rows;
    gint offset_x, offset_y;
    guint buf_size;
    guchar* buf;

    width  = gimp_image_width(image_id);
    height = gimp_image_height(image_id);
    
    drawable = gimp_drawable_get(mask_id);
    gimp_drawable_offsets(mask_id, &offset_x, &offset_y);

    cols = MIN((gint)drawable->width, width - offset_x);
    rows = MIN((gint)drawable->height, height - offset_y);
    
    buf_size = cols * rows;
    buf = (guchar*)g_malloc(buf_size);

    gimp_pixel_rgn_init(&pixel_rgn, drawable, offset_x, offset_y, cols, rows,
                        FALSE, FALSE);
    gimp_pixel_rgn_get_rect(&pixel_rgn, buf, offset_x, offset_y, cols, rows);
    gimp_drawable_detach(drawable);

    {
        gint dest_line_bit;
        gint dest_size_4bit;
        gint dest_size;
        guchar* dest;

        dest_line_bit  = ((width + 3) / 4) * 4;
        dest_size_4bit = dest_line_bit * height / 4 ;
        dest_size      = (dest_line_bit * height + 7) / 8;

        dest = (guchar*)g_malloc(dest_size);
        memset(dest, 0xFF, dest_size);

        {
            gint x, y;
            for (y = 0; y < rows; y++) {
                for (x = 0; x < cols; x++) {
                    gint x2, y2;
                    y2 = y + offset_y;
                    x2 = x + offset_x;
                    set_nth_bit(dest,
                                dest_line_bit * (height - 1 - y2) + x2,
                                (buf[y * height + x] > 127) ? 1 : 0);
                }
            }
        }

        {
            gint i = 0;
            gint state = -1;

            while (i < dest_size_4bit) {
                guchar value;
                guchar len;
                value = get_nth_4bit(dest, i++);
                len   = 1;
                while (i < dest_size_4bit) {
                    if (get_nth_4bit(dest, i) == value) {
                        len++;
                        i++;
                    } else {
                        break;
                    }
                }
                if (len == 1) {
                    write_4bit(value, &state, stream);
                } else {
                    write_4bit(value, &state, stream);
                    write_4bit(value, &state, stream);
                    write_4bit(len && 0x0F, &state, stream);
                    write_4bit(len >> 4, &state, stream);
                }
            }
            if (state != -1)
                write_4bit(0, &state, stream);
        }

        g_free(dest);
    }

    g_free(buf);
}


static
void write_additional_data_field(gconstpointer data, guint32 size, gchar* ID, FILE* stream)
{
    guint32 size_le;
    size_le = GINT32_TO_LE(size);
    fwrite(ID, 4, 1, stream);
    fwrite(&size_le, sizeof(guint32), 1, stream);
    if (size > 0)
        fwrite(data, size, 1, stream);
}


/*
 * GUIDE ==> GRID
 */

static
gint guide_entry_compare(gconstpointer a, gconstpointer b)
{
    return GPOINTER_TO_UINT(a) - GPOINTER_TO_UINT(b);
}

static
gboolean write_guide_as_grid(guint32 imageID, FILE* stream)
{
    GList *h_guides, *v_guides;
    gint32 guideID;
    DPGRID_HEADER header;
    guint x_grid_count, y_grid_count;
    guint i;

    h_guides = NULL;
    v_guides = NULL;

    guideID = gimp_image_find_next_guide(imageID, 0);
    while ((guideID != 0) && (guideID != -1)) {
        gint data;
        data = gimp_image_get_guide_position(imageID, guideID);
        if (data == -1)
            break;
        switch (gimp_image_get_guide_orientation(imageID, guideID)) {
        case GIMP_HORIZONTAL:
            h_guides = g_list_append(h_guides, GUINT_TO_POINTER(data));
            break;
        case GIMP_VERTICAL:
            v_guides = g_list_append(v_guides, GUINT_TO_POINTER(data));
            break;
        default:
            break;
        }
        guideID = gimp_image_find_next_guide(imageID, guideID);
    }

    h_guides = g_list_sort(g_list_first(h_guides), guide_entry_compare);
    v_guides = g_list_sort(g_list_first(v_guides), guide_entry_compare);

    x_grid_count = g_list_length(v_guides);
    y_grid_count = g_list_length(h_guides);

    fwrite("GRID", 4, 1, stream);
    {
        guint32 size;
        size = sizeof(header) + sizeof(guint16) * (x_grid_count + y_grid_count);
        size = GUINT32_TO_LE(size);
        fwrite(&size, sizeof(size), 1, stream);
    }

    memset(&header, 0, sizeof(header));
    sprintf(header.m_Comment,
            "DGR - D-Pixed grid file.\n"
            "Written by GIMP/%d.%d.%d with DPX-plugin/%s\n\n",
            gimp_major_version, gimp_minor_version, gimp_micro_version,
            VERSION);
    header.m_Width  = GUINT32_TO_LE(gimp_image_width(imageID));
    header.m_Height = GUINT32_TO_LE(gimp_image_height(imageID));
    header.m_XGrid  = GUINT32_TO_LE(x_grid_count);
    header.m_YGrid  = GUINT32_TO_LE(y_grid_count);
    header.m_Flags  = GUINT32_TO_LE(0);
    fwrite(&header, sizeof(header), 1, stream);

    for (i = 0; i < x_grid_count; i++) {
        guint16 tmp;
        tmp = GPOINTER_TO_UINT(g_list_nth_data(v_guides, i));
        tmp = GUINT16_TO_LE(tmp);
        fwrite(&tmp, sizeof(guint16), 1, stream);
    }
    for (i = 0; i < y_grid_count; i++) {
        guint16 tmp;
        tmp = GPOINTER_TO_UINT(g_list_nth_data(h_guides, i));
        tmp = GUINT16_TO_LE(tmp);
        fwrite(&tmp, sizeof(guint16), 1, stream);
    }

    g_list_free(h_guides);
    g_list_free(v_guides);

    return TRUE;
}
