/*
 * D-Pixed palette file converter for GIMP
 *
 * Copyright (C) 1999, 2000  Masahiro SAKAI
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "dpixplugins-conf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include "getopt.h"
#include "dpform.h"
#include "intl.h"


static
gboolean convert(const char* filename, FILE* input, FILE* output)
{
    gchar  str[512];
    gchar* tok;
    gint   r, g, b;
    gint   linenum;
    GList* palette;
    
    r = g = b = 0;

    linenum = 0;
    palette = NULL;

    if (!fgets (str, 512, input)) {
        g_message (_("Loading palette %s:\nRead error"), filename);
        return FALSE;
    }
    if (strcmp(str, "GIMP Palette\n") && strcmp(str, "GIMP Palette\r\n")) {
        g_message (_("Loading palette %s:\n"
                     "Corrupt palette: missing magic header"), filename);
        return FALSE;
    }

    linenum++;

    while (!feof (input)) {
        if (!fgets (str, 512, input)) {
            if (feof (input))
                break;
            g_message (_("Loading palette %s (line %d):\nRead error"),
                       filename, linenum);
            return FALSE;
	}
        
        linenum++;

        if (str[0] != '#') {
            tok = strtok (str, " \t");
            if (tok)
                r = atoi (tok);
            else
                /* maybe we should just abort? */
                g_message (_("Loading palette %s (line %d):\n"
                             "Missing RED component"), filename, linenum);

            tok = strtok (NULL, " \t");
            if (tok)
                g = atoi (tok);
            else
                g_message (_("Loading palette %s (line %d):\n"
                             "Missing GREEN component"), filename, linenum);

            tok = strtok (NULL, " \t");
            if (tok)
                b = atoi (tok);
            else
                g_message (_("Loading palette %s (line %d):\n"
                             "Missing BLUE component"), filename, linenum);

            /* optional name */
            tok = strtok (NULL, "\n");

            if (r < 0 || r > 255 ||
                g < 0 || g > 255 ||
                b < 0 || b > 255)
                g_message (_("Loading palette %s (line %d):\n"
                             "RGB value out of range"), filename, linenum);

            {
                DP_RGBQuad* rgb = g_new(DP_RGBQuad, 1);
                rgb->rgbBlue     = b;
                rgb->rgbGreen    = g;
                rgb->rgbRed      = r;
                rgb->rgbReserved = 0;
                palette = g_list_append(palette, rgb);
            }
	}
    }

    {
        gint palette_count = g_list_length(palette);
        DPIXED_FILEPALETTEHEADER header;
        int i;

        memset(&header, 0, sizeof(header));
        sprintf(header.fileDescription,
                "PF\n"
                "DPF - D-Pixed palette file.\n"
                "Written by DPF-plugin/%s\n\n",
                VERSION);
        header.version  = GUINT32_TO_LE(0x00000100);
        header.size     = GUINT32_TO_LE(palette_count);
        fwrite(&header, sizeof(header), 1, output);

        for (i = 0; i < palette_count; i++)
            fwrite(g_list_nth_data(palette, i), sizeof(DP_RGBQuad), 1, output);
    }

    g_list_free(palette);

    return TRUE;
}


/*
 *  Help and version information
 */
void usage (const char* progname) {
    fprintf(stderr, _("Usage: %s -o output,dpf filename\n"), progname);
}

const char* const help_message[] = {
    N_(" -o, --output=FILE              specify output file name as FILE"),
    N_(" -h, --help                     display this help message."),
    N_(" -v, --version                  display the version."),
    NULL
};

void
help(const char* progname)
{
    const char* const* p = help_message;

    fprintf(stderr, _("help: %s\n\n"), progname);
    usage(progname);
    while (*p) fprintf(stderr, "%s\n", _(*p++));
}

void version(const char* progname)
{
    fprintf(stderr, _("version of %s: %s\n"), progname, VERSION);
}


int main(int argc, char* argv[])
{
    static struct option longopts[] = {
        /* These options set a flag. */
        {"help",    0, NULL, 'h'},
        {"version", 0, NULL, 'v'},
        {"output",  1, NULL, 'o'},
        {0, 0, 0, 0}
    };
    int optc;
    char* progname = "gimppal2dpf";
    /* char *progname = argv[0]; */
    char* output_filename = NULL;
    char* input_filename = NULL;

    while ((optc = getopt_long(argc, argv, "hvo:", longopts, NULL)) != EOF) {
        switch (optc) {
        case 'h':
            help(progname);
            exit(EXIT_SUCCESS);
            break;
        case 'v':
            version(progname);
            exit(EXIT_SUCCESS);
            break;
        case 'o':
            output_filename = optarg;
            break;
        default:
            usage(progname);
            exit(EXIT_FAILURE);
        }
    }
    
    /* Test whether we have a palette file name as argument.  */
    if (optind >= argc) {
        fprintf(stderr, _("no input file given"));
        exit(EXIT_FAILURE);
    }
    input_filename = argv[optind];
    
    {
        FILE* input  = NULL;
        FILE* output = NULL;
        int result   = EXIT_FAILURE;

        if (strcmp(input_filename, "-") == 0)
            input = stdin;
        else
            input = fopen(input_filename, "rt");
        
        if (!input) {
            fprintf(stderr, _("failed in opening input file\n"));
            goto finally;
        }

        if (strcmp(output_filename, "-") == 0)
            output = stdout;
        else
            output = fopen(output_filename, "wb");

        if (!output) {
            fprintf(stderr, _("failed in opening output file\n"));
            goto finally;
        }

        if (!convert(input_filename, input, output))
            goto finally;

        result = EXIT_SUCCESS;

    finally:
        if (output && output != stdout)
            fclose(output);
        if (input && input != stdin)
            fclose(input);

        return result;
    }
}
